# @file dsc.py
# Data model for the EDK II DSC
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

# There will be some overlap between the objects for DSC files and FDF files
import collections
DEFAULT_SECTION_TYPE = "COMMON"


class dsc_set(set):

    def __init__(self, allowed_classes=None):
        if allowed_classes is None:
            allowed_classes = []
        self._allowed_classes = set(allowed_classes)

    def add(self, item):
        if len(self._allowed_classes) > 0 and type(item) not in self._allowed_classes:
            raise ValueError(f"Cannot add {type(item)} to restricted set: {self._allowed_classes}")
        if item in self:
            super().discard(item)
            # NEXTVER: get add the old_item to item
        super().add(item)


class dsc_dict(collections.OrderedDict):
    ''' A dictionary that allows specific classes as headers and sections '''

    def __init__(self, allowed_key_classes=None, allowed_value_classes=None):
        if allowed_key_classes is None:
            allowed_key_classes = []
        if allowed_value_classes is None:
            allowed_value_classes = []
        self._allowed_key_classes = set(allowed_key_classes)
        self._allowed_value_classes = set(allowed_value_classes)

    def __missing__(self, key):
        if len(self._allowed_value_classes) > 0:  # if we have specified allowed value classes, make a new dsc_set
            self[key] = dsc_set(self._allowed_value_classes)
            return self[key]
        raise KeyError(key)

    def __setitem__(self, key, val):
        if len(self._allowed_key_classes) > 0 and type(key) not in self._allowed_key_classes:
            raise ValueError(f"Cannot add {type(key)} to restricted set: {self._allowed_key_classes}")

        if len(self._allowed_value_classes) > 0:
            if type(val) == set and len(val) == 0:  # if it's an empty set, convert it to a dsc_set
                val = dsc_set(allowed_classes=self._allowed_value_classes)
            if type(val) == dsc_set:
                if val._allowed_classes != self._allowed_value_classes:
                    raise ValueError(
                        f"Cannot add set:{val._allowed_classes} to restricted dict: {self._allowed_value_classes}")
            elif type(val) not in self._allowed_value_classes:
                raise ValueError(f"Cannot add {type(val)} to restricted dict: {self._allowed_value_classes}")
        if key in self:
            # NEXTVER merge these together?
            raise ValueError(f"Cannot add section {key} since it already exists")
        dict.__setitem__(self, key, val)


class dsc:
    def __init__(self, file_path=None):
        # The EDK2 path to this particular DSC, if is is None it means it was created from a stream and has no file
        self.file_path = file_path
        # parameters added for clarity
        self.skus = dsc_set(allowed_classes=[definition, sku_id])  # this is a set of SKUs
        self.components = dsc_dict(allowed_key_classes=[dsc_section_type, ],
                                   allowed_value_classes=[component, definition])
        self.libraries = dsc_dict(allowed_key_classes=[dsc_section_type, ],
                                  allowed_value_classes=[library, definition])
        self.library_classes = dsc_dict(allowed_key_classes=[dsc_section_type, ],
                                        allowed_value_classes=[library_class, definition])
        self.build_options = dsc_dict(allowed_key_classes=[dsc_buildoption_section_type, ],
                                      allowed_value_classes=[build_option, definition])
        self.pcds = dsc_dict(allowed_key_classes=[dsc_pcd_section_type, ],
                             allowed_value_classes=[pcd, pcd_typed, pcd_variable])
        self.defines = dsc_set(allowed_classes=[definition, ])
        self.default_stores = dsc_set(allowed_classes=[definition, default_store])

        # NEXTVER: should we populate the default information into the DSC object?
        # FOR EXAMPLE: default_stores and skus

    def __eq__(self, other):
        ''' This doesn't check for a perfect copy of everything '''
        ''' this is mainly focused on does it define the same things '''
        if type(other) is not dsc:
            return False
        if other.skus != self.skus:
            return False
        if other.components != self.components:
            return False
        if other.library_classes != self.library_classes:
            return False
        if other.build_options != self.build_options:
            return False
        if other.pcds != self.pcds:
            return False
        if other.defines != self.defines:
            return False
        return True


class dsc_section_type:
    dsc_module_types = ["COMMON", "BASE", "SEC", "PEI_CORE", "PEIM", "DXE_CORE",
                        "DXE_DRIVER", "DXE_RUNTIME_DRIVER", "DXE_SAL_DRIVER",
                        "DXE_SMM_DRIVER", "SMM_CORE", "UEFI_DRIVER",
                        "UEFI_APPLICATION", "USER_DEFINED"]

    def __init__(self, arch="common", module_type="common"):
        self.arch = arch.upper().strip()
        self.module_type = module_type.upper().strip()
        if not dsc_section_type.IsValidModuleType(self.module_type):
            raise ValueError(f"{module_type} is not a proper module type for dsc section")

    def __hash__(self):
        arch = "*" if (self.arch == "COMMON" or self.arch == "DEFAULT") else self.arch
        return hash((arch, self.module_type))

    def __eq__(self, other):
        if type(other) is not dsc_section_type:
            return False
        arch = "*" if (self.arch == "COMMON" or self.arch == "DEFAULT") else self.arch
        arch2 = "*" if (other.arch == "COMMON" or other.arch == "DEFAULT") else other.arch
        return self.module_type == other.module_type or arch == arch2

    def __repr__(self):
        attributes = f".{self.arch}.{self.module_type}"
        return attributes

    @classmethod
    def IsValidModuleType(cls, name):
        return name in cls.dsc_module_types


class dsc_buildoption_section_type(dsc_section_type):
    '''
    [BuildOptions.$(arch).CodeBase.Edk2ModuleType]
    [BuildOptions.$(arch).CodeBase]
    [BuildOptions.common.CodeBase]
    [BuildOptions.$(arch)]
    [BuildOptions.common]
    [BuildOptions]
    '''

    def __init__(self, arch="common", codebase="common", module_type="common"):
        super().__init__(arch, module_type)
        self.codebase = codebase.upper().strip()
        if not self.IsValidCodeBase(self.codebase):
            raise ValueError(f"{codebase} is not a valid codebase type")

    @classmethod
    def IsValidCodeBase(cls, codebase):
        return codebase in ["COMMON", "EDK", "EDKII"]

    def __hash__(self):
        return hash((super().__hash__(), self.codebase))

    def __eq__(self, other):
        if type(other) is not dsc_buildoption_section_type:
            return False
        if not (other.module_type == self.module_type and other.codebase == self.codebase):
            return False
        return other.module_type == self.module_type

    def __repr__(self):
        return f"{self.arch}.{self.codebase}.{self.module_type}"


dsc_pcd_types = ["FEATUREFLAG", "PATCHABLEINMODULE", "FIXEDATBUILD", "DYNAMIC", "DYNAMICEX",
                 "DYNAMICDEFAULT", "DYNAMICHII", "DYNAMICVPD", "DYNAMICEXHII", "DYNAMICEXVPD"]


class dsc_pcd_section_type():
    def __init__(self, pcdtype, arch="common", sku="DEFAULT", store=None):
        # if store is none, then we don't have anything done
        self.arch = arch.upper().strip()
        self.pcd_type = pcdtype.upper().strip()
        self.default_store = None if store is None else store.strip()
        if self.pcd_type not in dsc_pcd_types:
            raise ValueError(f"{pcdtype} is not a proper PCD type")
        if not self.pcd_type.endswith("HII") and self.default_store is not None:
            raise ValueError(f"{pcdtype} does not allow for a store to be specified")
        self.sku = sku.upper()

    def __hash__(self):
        return hash((self.arch, self.pcd_type))

    def __repr__(self):
        store = "" if self.default_store is None else f".{self.default_store}"
        return f"Pcds{self.pcd_type}.{self.arch}.{self.sku}{store}"

    def __eq__(self, other):
        if type(other) is not dsc_pcd_section_type:
            return False
        return self.pcd_type == other.pcd_type and self.arch == other.arch


class dsc_pcd_component_type(dsc_pcd_section_type):
    ''' This class is uses to define the PCD type inside a component '''

    def __init__(self, pcdtype):
        super().__init__(pcdtype)

    def __repr__(self):
        return f"Pcds{self.pcd_type}"

    def __hash__(self):
        return hash(self.pcd_type)

    def __eq__(self, other):
        if type(other) is not dsc_pcd_component_type:
            return False
        return self.pcd_type == other.pcd_type


class sku_id:
    ''' contains the data for a sku '''

    def __init__(self, id=0, name="DEFAULT", parent="DEFAULT", source_info=None):
        self.id = id
        self.name = name
        self.parent = parent  # the default parent is default
        self.source_info = source_info

    def __eq__(self, other):
        if type(other) is not sku_id:
            return False
        return self.id == other.id or self.name == other.name

    def __hash__(self):
        # we return zero because we want all the skus to hash to the same bucket
        # this won't be performant for large numbers of skus, which hopefully won't happen
        # we instead rely on __eq__ since we want to collide on two different attributes
        # since we want to make sure names and id's are unique
        return 0

    def __repr__(self):
        return f"{self.id}|{self.name}|{self.parent}"


class component:
    ''' Contains the data for a component for the EDK build system to build '''

    def __init__(self, inf, source_info=None):

        self.library_classes = set()  # a list of libraries that this component uses
        self.pcds = {}  # a dictionary of PCD's that are keyed by dsc_pcd_component_type, they are sets
        self.defines = set()  # a set of defines
        self.build_options = set()  # a set of build options for this component
        self.inf = inf  # the EDK2 relative path to the source INF
        self.source_info = source_info

    def __eq__(self, other):
        if (type(other) is not component):
            return False
        return self.inf == other.inf  # NEXTVER: should this be case insensitive?

    def __hash__(self):
        return hash(self.inf)

    def __repr__(self):
        source = str(self.source_info) if source_info is not None else ""
        return f"{self.inf} @ {source}"


class definition:
    ''' contains the information on a definition. '''

    def __init__(self, name, value, local=False, source_info=None):
        ''' Local means DEFINE is in front and is localized to that particular section'''
        self.name = name
        self.value = value
        self.local = local
        self.source_info = source_info

    def __repr__(self):
        string = ""
        if self.local:
            string = "DEFINE "
        string += f"{self.name} = {self.value} @ {self.source_info}"
        return string

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if (type(other) is not definition):
            return False
        return other.name == self.name


class library:
    ''' Contains the data for a specific EDK library'''

    def __init__(self, inf: str, source_info=None):
        self.inf = inf
        self.source_info = source_info

    def __eq__(self, other):
        if type(other) is not library:
            return False
        return self.inf == other.inf

    def __hash__(self):
        return hash(self.inf)  # NEXTVER how to figure out if they hash to the same spot?

    def __repr__(self):
        return f"{self.inf} @ {self.source_info}"


class library_class:
    ''' Contains the data for a specific EDK2 library class'''

    def __init__(self, libraryclass: str, inf: str, source_info=None):
        self.libraryclass = libraryclass
        self.inf = inf
        self.source_info = source_info

    def __eq__(self, other):
        if (type(other) is not library_class):
            return False
        # if they're both null
        if self.libraryclass.lower() == "null" and other.libraryclass.lower() == "null":
            return self.inf == other.inf
        return self.libraryclass.lower() == other.libraryclass.lower()

    def __hash__(self):
        # if we're a null lib, we want the hash to be based on the inf path
        if (self.libraryclass.lower() == "null"):
            return hash(self.inf)
        else:
            return hash(self.libraryclass)

    def __repr__(self):
        return f"{self.libraryclass}|{self.inf} @ {self.source_info}"


class pcd:
    ''' Contains the data for a specific pcd '''
    ''' PcdTokenSpaceGuidCName.PcdCName|Value '''

    def __init__(self, namespace, name, value, source_info=None):
        self.namespace = namespace
        self.name = name
        self.value = value
        self.source_info = source_info

    def __eq__(self, other):
        if not issubclass(other.__class__, pcd):
            return False
        return self.namespace == other.namespace and self.name == other.name

    def __hash__(self):
        return hash(f"{self.namespace}.{self.name}")

    def __repr__(self):
        return f"{self.namespace}.{self.name} = {self.value} @ {self.source_info}"


class pcd_typed(pcd):
    ''' PcdTokenSpaceGuidCName.PcdCName|Value[|DatumType[|MaximumDatumSize]] '''

    def __init__(self, namespace, name, value, datum_type, max_size=0, source_info=None):
        super().__init__(namespace, name, value, source_info)
        self.datum_type = datum_type
        self.max_size = int(max_size)

    def __repr__(self):
        return f"{self.namespace}.{self.name} = {self.value} |{self.datum_type}|{self.max_size} @ {self.source_info}"


pcd_variable_attributes = ["NV", "BS", "RT", "RO"]


class pcd_variable(pcd):
    ''' PcdTokenSpaceGuidCName.PcdCName|VariableName|VariableGuid|VariableOffset[|HiiDefaultValue[|HiiAttribute]] '''

    def __init__(self, namespace, name, var_name, var_guid, var_offset,
                 default=None, attributes=None, source_info=None):
        super().__init__(namespace, name, "", source_info)
        if attributes is None:
            attributes = []
        self.var_name = var_name
        self.var_guid = var_guid
        self.var_offset = var_offset
        self.default = default

        if type(attributes) is str:
            attributes = attributes.split(",")
        attributes = [str(x).upper().strip() for x in attributes]
        if any([x not in pcd_variable_attributes for x in attributes]):
            raise ValueError(f"Invalid PcdHiiAttribute values: {attributes}")
        self.attributes = attributes

    def __repr__(self):
        pcd_data = f"{self.var_guid}|{self.var_offset}|{self.default}|{self.attributes}"
        return f"{self.namespace}.{self.name} = {self.var_name} |{pcd_data} @ {self.source_info}"


class build_option:
    ''' Contains the data for a build option '''
    ''' EX: MSFT:*_*_*_CC_FLAGS = /D MDEPKG_NDEBUG '''
    # {FAMILY}:{TARGET}_{TAGNAME}_{ARCH}_{TOOLCODE}_{ATTRIBUTE}

    def __init__(self, tool_code, attribute, data, target="*", tagname="*",
                 arch="*", family=None, replace=False, source_info=None):
        """
        tool_code - The tool code must be one of the defined tool codes in the Conf/tools_def.txt file.
         The flags defined in this section are appended to flags defined in the tools_def.txt file
         for individual tools.
        attribute - for example flags, d_path, path
        data - the actual flags or path you want to set
        target - DEBUG, RELEASE, or other
        tagname - the tool chain tag
        arch - ARM, AARCH64, IA32, X64, etc
        family - Conf/tools_def.txt defines FAMILY as one of MSFT, INTEL or GCC.
         Typically, this field is used to help the build tools determine whether the line
         is used for Microsoft style Makefiles or the GNU style Makefile
        replace - whether or not this replaces the default from tools_def,
         if this is false, we append
        """
        self.family = family
        self.target = target
        self.tagname = tagname
        self.arch = arch
        self.tool_code = tool_code
        self.attribute = attribute
        self.replace = replace
        self.data = data
        self.source_info = source_info

    def __eq__(self, other):
        if (type(other) is not build_option):
            return False
        if self.family != other.family:
            return False
        if self.target != other.target:
            return False
        if self.tagname != other.tagname:
            return False
        if self.arch != other.arch:
            return False
        if self.tool_code != other.tool_code:
            return False
        if self.attribute != other.attribute:
            return False
        return True

    def __hash__(self):
        return hash(self.__repr__(False))

    def __repr__(self, include_data=True):
        rep = "" if self.family is None else f"{self.family}:"
        rep += "_".join((self.target, self.tagname, self.arch, self.tool_code, self.attribute))
        if include_data:
            rep += f"= {self.data}"
        return rep


class default_store:
    ''' contains the information on a default store. '''
    ''' 0 | Standard        # UEFI Standard default '''

    def __init__(self, index=0, value="Standard", source_info=None):
        ''' Local means DEFINE is in front and is localized to that particular section'''
        self.index = int(index)
        self.value = value
        self.source_info = source_info

    def __repr__(self):
        return f"{self.index} | {self.value}"

    def __hash__(self):
        return hash(self.index)

    def __eq__(self, other):
        if (type(other) is not default_store):
            return False
        return other.index == self.index


class source_info:
    def __init__(self, file: str, lineno: int = None):
        self.file = file
        self.lineno = lineno

    def __repr__(self):
        if self.lineno is None:
            return self.file
        return f"{self.file}:{self.lineno}"



# @file
# Code to test UEFI MultiPhase module
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
from edk2toollib.uefi.uefi_multi_phase import (EfiVariableAttributes,
                                               EFI_VARIABLE_NON_VOLATILE, EFI_VARIABLE_RUNTIME_ACCESS,
                                               EFI_VARIABLE_BOOTSERVICE_ACCESS)


class TestUefiMultiphase (unittest.TestCase):

    def test_StringConversion(self):
        attr = EfiVariableAttributes(EFI_VARIABLE_NON_VOLATILE
                                     | EFI_VARIABLE_RUNTIME_ACCESS | EFI_VARIABLE_BOOTSERVICE_ACCESS)
        string = str(attr)

        self.assertTrue("EFI_VARIABLE_RUNTIME_ACCESS" in string)
        self.assertTrue("EFI_VARIABLE_NON_VOLATILE" in string)
        self.assertTrue("EFI_VARIABLE_BOOTSERVICE_ACCESS" in string)



# @file dsc_translator_test.py
# Tests for the translator for the EDK II DSC data model object
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
import os
import tempfile
from edk2toollib.uefi.edk2.parsers.dsc_parser import DscParser
# from edk2toollib.uefi.edk2.build_objects.dsc import dsc
from edk2toollib.uefi.edk2.build_objects.dsc_translator import DscTranslator


class TestDscTranslator(unittest.TestCase):
    test_dsc = """
## @file
# EFI/Framework Emulation Platform with UEFI HII interface supported.
#
# The Emulation Platform can be used to debug individual modules, prior
# to creating a real platform. This also provides an example for how
# an DSC is created.
#
# Copyright (c) 2006 - 2012, Intel Corporation. All rights reserved.<BR>
#
# This program and the accompanying materials are licensed and made
# available under the terms and conditions of the BSD License which
# accompanies this distribution.
# The full text of the license may be found at:
# http://opensource.org/licenses/bsd-license.php
#
# THE PROGRAM IS DISTRIBUTED UNDER THE BSD LICENSE ON AN "AS IS"
# BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS OF ANY KIND, EITHER
# EXPRESS OR IMPLIED.
#
##
########################################################################
#
# Defines Section - statements that will be processed to create a
# Makefile.
#
########################################################################
[Defines]
  PLATFORM_NAME           = NT32
  PLATFORM_GUID           = EB216561-961F-47EE-9EF9-CA426EF547C2
  PLATFORM_VERSION        = 0.5
  DSC_SPECIFICATION       = 0x0001001C
  OUTPUT_DIRECTORY        = Build/NT32
  SUPPORTED_ARCHITECTURES = IA32
  BUILD_TARGETS           = DEBUG|RELEASE
  SKUID_IDENTIFIER        = DEFAULT
  FLASH_DEFINITION        = Nt32Pkg/Nt32Pkg.fdf
  #
  # Defines for default states. These can be changed on the command
  # line.
  # -D FLAG=VALUE
  #
!ifndef SECURE_BOOT_ENABLE
  DEFINE SECURE_BOOT_ENABLE = FALSE
!endif
  DEFINE SECRET_VAR = FALSE

########################################################################
#
# SKU Identification section - list of all SKU IDs supported by this
# Platform.
#
########################################################################
[SkuIds]
  0|DEFAULT # The entry: 0|DEFAULT is reserved and always required.
  3|JOEY|DEFAULT
  5|JOEY2|JOEY

########################################################################
#
# Library Class section - list of all Library Classes needed by this
# Platform.
#
########################################################################
[LibraryClasses]
  DEFINE SECRET_VAR = TRUE
  #
  # Entry point
  #
  PeiCoreEntryPoint|MdePkg/Library/PeiCoreEntryPoint/PeiCoreEntryPoint.inf
  PeimEntryPoint|MdePkg/Library/PeimEntryPoint/PeimEntryPoint.inf
  DxeCoreEntryPoint|MdePkg/Library/DxeCoreEntryPoint/DxeCoreEntryPoint.inf
  UefiDriverEntryPoint|MdePkg/Library/UefiDriverEntryPoint/UefiDriverEntryPoint.inf
  UefiApplicationEntryPoint|MdePkg/Library/UefiApplicationEntryPoint/UefiApplicationEntryPoint.inf
  #
  # Basic
  #
  BaseLib|MdePkg/Library/BaseLib/BaseLib.inf
  SynchronizationLib|MdePkg/Library/BaseSynchronizationLib/BaseSynchronizationLib.inf
  PrintLib|MdePkg/Library/BasePrintLib/BasePrintLib.inf
  CpuLib|MdePkg/Library/BaseCpuLib/BaseCpuLib.inf
  IoLib|MdePkg/Library/BaseIoLibIntrinsic/BaseIoLibIntrinsic.inf
  PciLib|MdePkg/Library/BasePciLibCf8/BasePciLibCf8.inf
  PciCf8Lib|MdePkg/Library/BasePciCf8Lib/BasePciCf8Lib.inf
  PciExpressLib|MdePkg/Library/BasePciExpressLib/BasePciExpressLib.inf
  CacheMaintenanceLib|MdePkg/Library/BaseCacheMaintenanceLib/BaseCacheMaintenanceLib.inf
  PeCoffLib|MdePkg/Library/BasePeCoffLib/BasePeCoffLib.inf
  PeCoffGetEntryPointLib|MdePkg/Library/BasePeCoffGetEntryPointLib/BasePeCoffGetEntryPointLib.inf
  #
  # UEFI & PI
  #
  UefiBootServicesTableLib|MdePkg/Library/UefiBootServicesTableLib/UefiBootServicesTableLib.inf
  UefiRuntimeServicesTableLib|MdePkg/Library/UefiRuntimeServicesTableLib/UefiRuntimeServicesTableLib.inf
  UefiRuntimeLib|MdePkg/Library/UefiRuntimeLib/UefiRuntimeLib.inf
  UefiLib|MdePkg/Library/UefiLib/UefiLib.inf
  UefiHiiServicesLib|MdeModulePkg/Library/UefiHiiServicesLib/UefiHiiServicesLib.inf
  HiiLib|MdeModulePkg/Library/UefiHiiLib/UefiHiiLib.inf
  DevicePathLib|MdePkg/Library/UefiDevicePathLib/UefiDevicePathLib.inf
  UefiDecompressLib|IntelFrameworkModulePkg/Library/BaseUefiTianoCustomDecompressLib/BaseUefiTianoCustomDecompressLib.inf
  PeiServicesTablePointerLib|MdePkg/Library/PeiServicesTablePointerLib/PeiServicesTablePointerLib.inf
  PeiServicesLib|MdePkg/Library/PeiServicesLib/PeiServicesLib.inf
  DxeServicesLib|MdePkg/Library/DxeServicesLib/DxeServicesLib.inf
  DxeServicesTableLib|MdePkg/Library/DxeServicesTableLib/DxeServicesTableLib.inf
  #
  # Generic Modules
  #
  UefiUsbLib|MdePkg/Library/UefiUsbLib/UefiUsbLib.inf
  UefiScsiLib|MdePkg/Library/UefiScsiLib/UefiScsiLib.inf
  NetLib|MdeModulePkg/Library/DxeNetLib/DxeNetLib.inf
  IpIoLib|MdeModulePkg/Library/DxeIpIoLib/DxeIpIoLib.inf
  UdpIoLib|MdeModulePkg/Library/DxeUdpIoLib/DxeUdpIoLib.inf
  DpcLib|MdeModulePkg/Library/DxeDpcLib/DxeDpcLib.inf
  OemHookStatusCodeLib|MdeModulePkg/Library/OemHookStatusCodeLibNull/OemHookStatusCodeLibNull.inf
  GenericBdsLib|IntelFrameworkModulePkg/Library/GenericBdsLib/GenericBdsLib.inf
  SecurityManagementLib|MdeModulePkg/Library/DxeSecurityManagementLib/DxeSecurityManagementLib.inf
  TimerLib|MdePkg/Library/BaseTimerLibNullTemplate/BaseTimerLibNullTemplate.inf
  SerialPortLib|MdePkg/Library/BaseSerialPortLibNull/BaseSerialPortLibNull.inf
  CapsuleLib|MdeModulePkg/Library/DxeCapsuleLibNull/DxeCapsuleLibNull.inf
  #
  # Platform
  #
  PlatformBdsLib|Nt32Pkg/Library/Nt32BdsLib/Nt32BdsLib.inf
  #
  # Misc
  #
  DebugLib|IntelFrameworkModulePkg/Library/PeiDxeDebugLibReportStatusCode/PeiDxeDebugLibReportStatusCode.inf
  DebugPrintErrorLevelLib|MdeModulePkg/Library/DxeDebugPrintErrorLevelLib/DxeDebugPrintErrorLevelLib.inf
  PerformanceLib|MdePkg/Library/BasePerformanceLibNull/BasePerformanceLibNull.inf
  DebugAgentLib|MdeModulePkg/Library/DebugAgentLibNull/DebugAgentLibNull.inf
  CpuExceptionHandlerLib|MdeModulePkg/Library/CpuExceptionHandlerLibNull/CpuExceptionHandlerLibNull.inf
!if $(SECURE_BOOT_ENABLE) == TRUE
  PlatformSecureLib|Nt32Pkg/Library/PlatformSecureLib/PlatformSecureLib.inf
  IntrinsicLib|CryptoPkg/Library/IntrinsicLib/IntrinsicLib.inf
  OpensslLib|CryptoPkg/Library/OpensslLib/OpensslLib.inf
!endif

[LibraryClasses.common.USER_DEFINED]
  DebugLib|MdePkg/Library/BaseDebugLibNull/BaseDebugLibNull.inf
  PeCoffExtraActionLib|MdePkg/Library/BasePeCoffExtraActionLibNull/BasePeCoffExtraActionLibNull.inf
  ReportStatusCodeLib|MdeModulePkg/Library/PeiReportStatusCodeLib/PeiReportStatusCodeLib.inf
  OemHookStatusCodeLib|Nt32Pkg/Library/PeiNt32OemHookStatusCodeLib/PeiNt32OemHookStatusCodeLib.inf
  MemoryAllocationLib|MdePkg/Library/PeiMemoryAllocationLib/PeiMemoryAllocationLib.inf
  PcdLib|MdePkg/Library/BasePcdLibNull/BasePcdLibNull.inf

[LibraryClasses.common.PEIM,LibraryClasses.common.PEI_CORE]
  #
  # PEI phase common
  #
  HobLib|MdePkg/Library/PeiHobLib/PeiHobLib.inf
  MemoryAllocationLib|MdePkg/Library/PeiMemoryAllocationLib/PeiMemoryAllocationLib.inf
  ReportStatusCodeLib|MdeModulePkg/Library/PeiReportStatusCodeLib/PeiReportStatusCodeLib.inf
  ExtractGuidedSectionLib|MdePkg/Library/PeiExtractGuidedSectionLib/PeiExtractGuidedSectionLib.inf
  BaseMemoryLib|MdePkg/Library/BaseMemoryLibOptPei/BaseMemoryLibOptPei.inf
  IoLib|MdePkg/Library/PeiIoLibCpuIo/PeiIoLibCpuIo.inf
  PeCoffGetEntryPointLib|Nt32Pkg/Library/Nt32PeiPeCoffGetEntryPointLib/Nt32PeiPeCoffGetEntryPointLib.inf
  PeCoffExtraActionLib|Nt32Pkg/Library/PeiNt32PeCoffExtraActionLib/PeiNt32PeCoffExtraActionLib.inf
  DebugPrintErrorLevelLib|MdePkg/Library/BaseDebugPrintErrorLevelLib/BaseDebugPrintErrorLevelLib.inf

[LibraryClasses.common.PEI_CORE]
  PcdLib|MdePkg/Library/BasePcdLibNull/BasePcdLibNull.inf
  OemHookStatusCodeLib|MdeModulePkg/Library/OemHookStatusCodeLibNull/OemHookStatusCodeLibNull.inf

[LibraryClasses.common.PEIM]
  PcdLib|MdePkg/Library/PeiPcdLib/PeiPcdLib.inf
  OemHookStatusCodeLib|Nt32Pkg/Library/PeiNt32OemHookStatusCodeLib/PeiNt32OemHookStatusCodeLib.inf
!if $(SECURE_BOOT_ENABLE) == TRUE
  BaseCryptLib|CryptoPkg/Library/BaseCryptLib/PeiCryptLib.inf
!endif

[LibraryClasses.common]
  #
  # DXE phase common
  #
  BaseMemoryLib|MdePkg/Library/BaseMemoryLibOptDxe/BaseMemoryLibOptDxe.inf
  HobLib|MdePkg/Library/DxeHobLib/DxeHobLib.inf
  PcdLib|MdePkg/Library/DxePcdLib/DxePcdLib.inf
  MemoryAllocationLib|MdePkg/Library/UefiMemoryAllocationLib/UefiMemoryAllocationLib.inf
  ReportStatusCodeLib|MdeModulePkg/Library/DxeReportStatusCodeLib/DxeReportStatusCodeLib.inf
  OemHookStatusCodeLib|Nt32Pkg/Library/DxeNt32OemHookStatusCodeLib/DxeNt32OemHookStatusCodeLib.inf
  PeCoffExtraActionLib|Nt32Pkg/Library/DxeNt32PeCoffExtraActionLib/DxeNt32PeCoffExtraActionLib.inf
  ExtractGuidedSectionLib|MdePkg/Library/DxeExtractGuidedSectionLib/DxeExtractGuidedSectionLib.inf
  WinNtLib|Nt32Pkg/Library/DxeWinNtLib/DxeWinNtLib.inf
!if $(SECURE_BOOT_ENABLE) == TRUE
  BaseCryptLib|CryptoPkg/Library/BaseCryptLib/BaseCryptLib.inf
!endif

[LibraryClasses.common.DXE_CORE]
  HobLib|MdePkg/Library/DxeCoreHobLib/DxeCoreHobLib.inf
  MemoryAllocationLib|MdeModulePkg/Library/DxeCoreMemoryAllocationLib/DxeCoreMemoryAllocationLib.inf
  PcdLib|MdePkg/Library/BasePcdLibNull/BasePcdLibNull.inf

[LibraryClasses.common.DXE_SMM_DRIVER]
  DebugLib|MdePkg/Library/BaseDebugLibNull/BaseDebugLibNull.inf

[LibraryClasses.common.UEFI_DRIVER]
  PcdLib|MdePkg/Library/BasePcdLibNull/BasePcdLibNull.inf

[LibraryClasses.common.UEFI_APPLICATION]
  PcdLib|MdePkg/Library/BasePcdLibNull/BasePcdLibNull.inf
  PrintLib|MdeModulePkg/Library/DxePrintLibPrint2Protocol/DxePrintLibPrint2Protocol.inf

[LibraryClasses.common.DXE_RUNTIME_DRIVER]
  #
  # Runtime
  #
!if $(SECURE_BOOT_ENABLE) == TRUE
  BaseCryptLib|CryptoPkg/Library/BaseCryptLib/RuntimeCryptLib.inf
!endif

########################################################################
#
# Pcd Section - list of all EDK II PCD Entries defined by this Platform
#
########################################################################
[PcdsFeatureFlag]
  gEfiMdeModulePkgTokenSpaceGuid.PcdStatusCodeUseSerial|FALSE
  gEfiMdeModulePkgTokenSpaceGuid.PcdDxeIplSwitchToLongMode|FALSE
  gEfiMdeModulePkgTokenSpaceGuid.PcdPeiCoreImageLoaderSearchTeSectionFirst|FALSE
  gEfiMdeModulePkgTokenSpaceGuid.PcdVariableCollectStatistics|TRUE

[PcdsFixedAtBuild]
  gEfiMdeModulePkgTokenSpaceGuid.PcdMaxSizeNonPopulateCapsule|0x0
  gEfiMdeModulePkgTokenSpaceGuid.PcdMaxSizePopulateCapsule|0x0
  gEfiMdePkgTokenSpaceGuid.PcdDebugPrintErrorLevel|0x80000040
  gEfiNt32PkgTokenSpaceGuid.PcdWinNtFirmwareFdSize|0x2a0000
  gEfiMdePkgTokenSpaceGuid.PcdDebugPropertyMask|0x1f
  gEfiNt32PkgTokenSpaceGuid.PcdWinNtFirmwareVolume|L"..FvNt32.fd"
  gEfiNt32PkgTokenSpaceGuid.PcdWinNtFirmwareBlockSize|0x10000
  gEfiMdePkgTokenSpaceGuid.PcdReportStatusCodePropertyMask|0x0f
  gEfiMdeModulePkgTokenSpaceGuid.PcdResetOnMemoryTypeInformationChange| FALSE
!if $(SECURE_BOOT_ENABLE) == TRUE
  gEfiMdeModulePkgTokenSpaceGuid.PcdMaxVariableSize|0x2000
!endif
!if $(SECURE_BOOT_ENABLE) == TRUE
  # override the default values from SecurityPkg to ensure images from
  # all sources are verified in secure boot
  gEfiSecurityPkgTokenSpaceGuid.PcdOptionRomImageVerificationPolicy|0x05
  gEfiSecurityPkgTokenSpaceGuid.PcdFixedMediaImageVerificationPolicy|0x05
  gEfiSecurityPkgTokenSpaceGuid.PcdRemovableMediaImageVerificationPolicy|0x05
!endif

[PcdsPatchableInModule.IA32, PcdsPatchableInModule.X64]
  gEfiNt32PkgTokenSpaceGuid.PcdWinNtGop|L"UGA Window 1!UGA Window 2"|VOID*|52

########################################################################
#
# Pcd Dynamic Section - list of all EDK II PCD Entries defined by this Platform
#
########################################################################
[PcdsDynamicDefault.common.DEFAULT]
  gEfiNt32PkgTokenSpaceGuid.PcdWinNtSerialPort|L"COM1!COM2"|VOID*|20
  gEfiNt32PkgTokenSpaceGuid.PcdWinNtFileSystem|L".!.....EdkShellBinPkgBinIa32Apps"|VOID*|106
  gEfiNt32PkgTokenSpaceGuid.PcdWinNtGop|L"UGA Window 1!UGA Window 2"|VOID*|52
  gEfiNt32PkgTokenSpaceGuid.PcdWinNtConsole|L"Bus Driver Console Window"|VOID*|52
  gEfiNt32PkgTokenSpaceGuid.PcdWinNtVirtualDisk|L"FW;40960;512"|VOID*|26
  gEfiNt32PkgTokenSpaceGuid.PcdWinNtMemorySize|L"64!64"|VOID*|12
  gEfiNt32PkgTokenSpaceGuid.PcdWinNtPhysicalDisk|L"a:RW;2880;512!d:RO;307200;2048!j:RW;262144;512"|VOID*|100
  gEfiNt32PkgTokenSpaceGuid.PcdWinNtUga|L"UGA Window 1!UGA Window 2"|VOID*|52
  gEfiMdeModulePkgTokenSpaceGuid.PcdFlashNvStorageFtwSpareBase|0
  gEfiMdeModulePkgTokenSpaceGuid.PcdFlashNvStorageFtwWorkingBase|0
  gEfiMdeModulePkgTokenSpaceGuid.PcdFlashNvStorageVariableBase|0

[PcdsDynamicHii.common.DEFAULT]
  gEfiIntelFrameworkModulePkgTokenSpaceGuid.PcdSetupConOutColumn|L"SetupConsoleConfig"|gEfiGlobalVariableGuid|0x0|80
  gEfiIntelFrameworkModulePkgTokenSpaceGuid.PcdSetupConOutRow|L"SetupConsoleConfig"|gEfiGlobalVariableGuid|0x4|25
  gEfiIntelFrameworkModulePkgTokenSpaceGuid.PcdPlatformBootTimeOut|L"Timeout"|gEfiGlobalVariableGuid|0x0|10
  gEfiIntelFrameworkModulePkgTokenSpaceGuid.PcdHardwareErrorRecordLevel|L"HwErrRecSupport"|gEfiGlobalVariableGuid|0x0|1
  gEfiMdeModulePkgTokenSpaceGuid.PcdValidRange|L"PcdValidRange"|gEfiGlobalVariableGuid|0x07|0|BS,RT,NV

########################################################################
#
# Components Section - list of the modules and components that will be
# processed by compilation tools and the EDK II
# tools to generate PE32/PE32+/Coff image files.
#
# Note: The EDK II DSC file is not used to specify how compiled binary
# images get placed into firmware volume images. This section is
# just a list of modules to compile from source into
# UEFI-compliant binaries.
# It is the FDF file that contains information on combining binary
# files into firmware volume images, whose concept is beyond UEFI
# and is described in PI specification.
# Binary modules do not need to be listed in this section, as they
# should be specified in the FDF file. For example: Shell binary
# (Shell_Full.efi), FAT binary (Fat.efi), Logo (Logo.bmp), and etc.
# There may also be modules listed in this section that are not
# required in the FDF file,
# When a module listed here is excluded from FDF file, then
# UEFI-compliant binary will be generated for it, but the binary
# will not be put into any firmware volume.
#
########################################################################
[Components.IA32]
  ##
  # SEC Phase modules
  ##
  Nt32Pkg/Sec/SecMain.inf
  ##
  # PEI Phase modules
  ##
  MdeModulePkg/Core/Pei/PeiMain.inf
  MdeModulePkg/Universal/PCD/Pei/Pcd.inf {
    <LibraryClasses>
      PcdLib|MdePkg/Library/BasePcdLibNull/BasePcdLibNull.inf
  }
  MdeModulePkg/Universal/ReportStatusCodeRouter/Pei/ReportStatusCodeRouterPei.inf
  MdeModulePkg/Universal/StatusCodeHandler/Pei/StatusCodeHandlerPei.inf
  Nt32Pkg/WinNtOemHookStatusCodeHandlerPei/WinNtOemHookStatusCodeHandlerPei.inf
  Nt32Pkg/BootModePei/BootModePei.inf
  Nt32Pkg/StallPei/StallPei.inf
  Nt32Pkg/WinNtFlashMapPei/WinNtFlashMapPei.inf
!if $(SECURE_BOOT_ENABLE) == TRUE
  SecurityPkg/VariableAuthenticated/Pei/VariablePei.inf
!else
  MdeModulePkg/Universal/Variable/Pei/VariablePei.inf
!endif

  Nt32Pkg/WinNtAutoScanPei/WinNtAutoScanPei.inf
  Nt32Pkg/WinNtFirmwareVolumePei/WinNtFirmwareVolumePei.inf
  Nt32Pkg/WinNtThunkPPIToProtocolPei/WinNtThunkPPIToProtocolPei.inf
  MdeModulePkg/Core/DxeIplPeim/DxeIpl.inf

[Components.X64]
  ##
  # DXE Phase modules
  ##
  MdeModulePkg/Core/Dxe/DxeMain.inf {
    <LibraryClasses>
      NULL| MdeModulePkg/Library/DxeCrc32GuidedSectionExtractLib/DxeCrc32GuidedSectionExtractLib.inf
    <BuildOptions>
      *_*_IA32_CC_FLAGS =
  }

  MdeModulePkg/Universal/PCD/Dxe/Pcd.inf {
    <LibraryClasses>
      PcdLib|MdePkg/Library/BasePcdLibNull/BasePcdLibNull.inf
  }
  Nt32Pkg/MetronomeDxe/MetronomeDxe.inf
  Nt32Pkg/RealTimeClockRuntimeDxe/RealTimeClockRuntimeDxe.inf
  Nt32Pkg/ResetRuntimeDxe/ResetRuntimeDxe.inf
  MdeModulePkg/Core/RuntimeDxe/RuntimeDxe.inf
  Nt32Pkg/FvbServicesRuntimeDxe/FvbServicesRuntimeDxe.inf
  MdeModulePkg/Universal/SecurityStubDxe/SecurityStubDxe.inf {
    <LibraryClasses>
      !if $(SECURE_BOOT_ENABLE) == TRUE
        NULL|SecurityPkg/Library/DxeImageVerificationLib/DxeImageVerificationLib.inf
      !endif
  }
  MdeModulePkg/Universal/SmbiosDxe/SmbiosDxe.inf
  MdeModulePkg/Universal/EbcDxe/EbcDxe.inf
  MdeModulePkg/Universal/MemoryTest/NullMemoryTestDxe/NullMemoryTestDxe.inf
  Nt32Pkg/WinNtThunkDxe/WinNtThunkDxe.inf
  Nt32Pkg/CpuRuntimeDxe/CpuRuntimeDxe.inf
  MdeModulePkg/Universal/FaultTolerantWriteDxe/FaultTolerantWriteDxe.inf
  Nt32Pkg/MiscSubClassPlatformDxe/MiscSubClassPlatformDxe.inf
  Nt32Pkg/TimerDxe/TimerDxe.inf
  MdeModulePkg/Universal/ReportStatusCodeRouter/RuntimeDxe/ReportStatusCodeRouterRuntimeDxe.inf
  MdeModulePkg/Universal/StatusCodeHandler/RuntimeDxe/StatusCodeHandlerRuntimeDxe.inf
  Nt32Pkg/WinNtOemHookStatusCodeHandlerDxe/WinNtOemHookStatusCodeHandlerDxe.inf
!if $(SECURE_BOOT_ENABLE) == TRUE
  SecurityPkg/VariableAuthenticated/RuntimeDxe/VariableRuntimeDxe.inf
  SecurityPkg/VariableAuthenticated/SecureBootConfigDxe/ SecureBootConfigDxe.inf
!else
  MdeModulePkg/Universal/Variable/RuntimeDxe/VariableRuntimeDxe.inf
!endif
  MdeModulePkg/Universal/WatchdogTimerDxe/WatchdogTimer.inf
  MdeModulePkg/Universal/MonotonicCounterRuntimeDxe/MonotonicCounterRuntimeDxe.inf
  MdeModulePkg/Universal/CapsuleRuntimeDxe/CapsuleRuntimeDxe.inf
  MdeModulePkg/Universal/Console/ConPlatformDxe/ConPlatformDxe.inf
  MdeModulePkg/Universal/Console/ConSplitterDxe/ConSplitterDxe.inf {
    <LibraryClasses>
      PcdLib|MdePkg/Library/DxePcdLib/DxePcdLib.inf
  }
  MdeModulePkg/Universal/Console/GraphicsConsoleDxe/GraphicsConsoleDxe.inf {
    <LibraryClasses>
      PcdLib|MdePkg/Library/DxePcdLib/DxePcdLib.inf
  }
  MdeModulePkg/Universal/Console/TerminalDxe/TerminalDxe.inf {
    <LibraryClasses>
      PcdLib|MdePkg/Library/DxePcdLib/DxePcdLib.inf
  }
  MdeModulePkg/Universal/DevicePathDxe/DevicePathDxe.inf
  MdeModulePkg/Universal/Disk/DiskIoDxe/DiskIoDxe.inf
  MdeModulePkg/Universal/Disk/PartitionDxe/PartitionDxe.inf
  MdeModulePkg/Universal/Disk/UnicodeCollation/EnglishDxe/EnglishDxe.inf
  MdeModulePkg/Bus/Pci/PciBusDxe/PciBusDxe.inf
  MdeModulePkg/Bus/Scsi/ScsiBusDxe/ScsiBusDxe.inf ## This driver follows UEFI
  ## specification definition
  MdeModulePkg/Bus/Scsi/ScsiDiskDxe/ScsiDiskDxe.inf ## This driver follows UEFI
  ## specification definition
  IntelFrameworkModulePkg/Bus/Pci/IdeBusDxe/IdeBusDxe.inf
  Nt32Pkg/WinNtBusDriverDxe/WinNtBusDriverDxe.inf {
    <LibraryClasses>
      PcdLib|MdePkg/Library/DxePcdLib/DxePcdLib.inf
  }
  Nt32Pkg/WinNtBlockIoDxe/WinNtBlockIoDxe.inf
  Nt32Pkg/WinNtSerialIoDxe/WinNtSerialIoDxe.inf
  Nt32Pkg/WinNtGopDxe/WinNtGopDxe.inf
  Nt32Pkg/WinNtSimpleFileSystemDxe/WinNtSimpleFileSystemDxe.inf
  MdeModulePkg/Application/HelloWorld/HelloWorld.inf
  #
  # Network stack drivers
  # To test network drivers, need network Io driver(SnpNt32Io.dll), please refer
  # to NETWORK-IO Subproject.
  #
  MdeModulePkg/Universal/Network/DpcDxe/DpcDxe.inf
  MdeModulePkg/Universal/Network/ArpDxe/ArpDxe.inf
  MdeModulePkg/Universal/Network/Dhcp4Dxe/Dhcp4Dxe.inf
  MdeModulePkg/Universal/Network/Ip4ConfigDxe/Ip4ConfigDxe.inf
  MdeModulePkg/Universal/Network/Ip4Dxe/Ip4Dxe.inf
  MdeModulePkg/Universal/Network/MnpDxe/MnpDxe.inf
  MdeModulePkg/Universal/Network/VlanConfigDxe/VlanConfigDxe.inf
  MdeModulePkg/Universal/Network/Mtftp4Dxe/Mtftp4Dxe.inf
  MdeModulePkg/Universal/Network/Tcp4Dxe/Tcp4Dxe.inf
  MdeModulePkg/Universal/Network/Udp4Dxe/Udp4Dxe.inf
  MdeModulePkg/Universal/Network/UefiPxeBcDxe/UefiPxeBcDxe.inf
  Nt32Pkg/SnpNt32Dxe/SnpNt32Dxe.inf
  MdeModulePkg/Universal/Network/IScsiDxe/IScsiDxe.inf
  IntelFrameworkModulePkg/Universal/BdsDxe/BdsDxe.inf
  MdeModulePkg/Universal/HiiDatabaseDxe/HiiDatabaseDxe.inf
  MdeModulePkg/Universal/SetupBrowserDxe/SetupBrowserDxe.inf
  MdeModulePkg/Universal/PrintDxe/PrintDxe.inf
  MdeModulePkg/Universal/DriverSampleDxe/DriverSampleDxe.inf {
    <LibraryClasses>
      PcdLib|MdePkg/Library/BasePcdLibNull/BasePcdLibNull.inf
    <PcdsFeatureFlag>
      gEfiMdeModulePkgTokenSpaceGuid.PcdStatusCodeUseSerial|TRUE
  }
  MdeModulePkg/Application/VariableInfo/VariableInfo.inf
  MdeModulePkg/Universal/PlatformDriOverrideDxe/PlatformDriOverrideDxe.inf

######################################################################## #
# BuildOptions Section - Define the module specific tool chain flags that
# should be used as the default flags for a
# module. These flags are appended to any
# standard flags that are defined by the build
# process. They can be applied for any modules or
# only those modules with the specific module
# style (EDK or EDKII) specified in [Components] # section.
#
########################################################################
[BuildOptions]
  DEBUG_*_IA32_DLINK_FLAGS = /BASE:0x10000 /ALIGN:4096 /FILEALIGN:4096 \\
                             /EXPORT:InitializeDriver=$(IMAGE_ENTRY_POINT) \\
                             /SUBSYSTEM:CONSOLE
  RELEASE_*_IA32_DLINK_FLAGS = /ALIGN:4096 /FILEALIGN:4096
  *_*_IA32_CC_FLAGS = /D EFI_SPECIFICATION_VERSION = 0x0002000A \\
                      /D TIANO_RELEASE_VERSION=0x00080006

"""

    def write_file(self, file_path, contents):
        temp_dir = tempfile.mkdtemp()
        file_path = os.path.join(temp_dir, file_path)
        f = open(file_path, "w")
        f.write(contents)
        f.close()
        return file_path

    def test_dsc_to_file_and_back_again(self):
        filepath = self.write_file("test.dsc", self.test_dsc)
        print(filepath)
        # parse the original DSC
        parser = DscParser()
        parser._ErrorLimit = 0
        # NEXTVER: actually parse the file
        dsc_obj = []  # parser.ParseFile(filepath)
        # Write out to disk
        test_path = os.path.join(os.path.dirname(filepath), "test2.dsc")
        DscTranslator.dsc_to_file(dsc_obj, test_path)
        # parse in the outputted DSC
        parser2 = DscParser()
        parser2._ErrorLimit = 0
        print(test_path)
        # NEXTVER: actually parse the file
        # dsc_obj2 = parser2.ParseFile(test_path)
        # self.assertNotEqual(dsc_obj, None)
        # self.assertNotEqual(dsc_obj2, None)
        # self.assertEqual(len(dsc_obj.defines), len(dsc_obj2.defines))
        # self.assertEqual(len(dsc_obj.library_classes), len(dsc_obj2.library_classes))
        # self.assertEqual(len(dsc_obj.components), len(dsc_obj2.components))
        # self.assertEqual(len(dsc_obj.build_options), len(dsc_obj2.build_options))
        # self.assertEqual(len(dsc_obj.pcds), len(dsc_obj2.pcds))
        # self.assertEqual(dsc_obj, dsc_obj2)



# @file
# Unit test harness for the VariablePolicy module/classes.
#
# Copyright (c) Microsoft Corporation
# SPDX-License-Identifier: BSD-2-Clause-Patent
##


import unittest
import uuid
from edk2toollib.uefi.edk2.variable_policy import VariableLockOnVarStatePolicy, VariablePolicyEntry


TEST_GUID_1 = uuid.UUID("48B5F961-3F7D-4B88-9BEE-D305ED8256DA")
TEST_GUID_2 = uuid.UUID("65D16747-FCBC-4FAE-A727-7B679A7B23F9")

TEST_POLICY_ENTRY = b''.fromhex("000001006A004600E222FFB0EA4A2547A6E55317FB8FD39C00000000FFFFFFFF000000000000000003AFAFAFC690F5ECF9F887438422486E3CCD8B2001AF45004F00440000004C0061007300740041007400740065006D00700074005300740061007400750073000000")                 # noqa
TEST_POLICY_ENTRY_BAD_VERSION = b''.fromhex("010001006A004600E222FFB0EA4A2547A6E55317FB8FD39C00000000FFFFFFFF000000000000000003AFAFAFC690F5ECF9F887438422486E3CCD8B2001AF45004F00440000004C0061007300740041007400740065006D00700074005300740061007400750073000000")     # noqa
TEST_POLICY_ENTRY_BAD_LOCK_TYPE = b''.fromhex("000001006A004600E222FFB0EA4A2547A6E55317FB8FD39C00000000FFFFFFFF000000000000000004AFAFAFC690F5ECF9F887438422486E3CCD8B2001AF45004F00440000004C0061007300740041007400740065006D00700074005300740061007400750073000000")   # noqa
TEST_POLICY_ENTRY_GUID = uuid.UUID("B0FF22E2-4AEA-4725-A6E5-5317FB8FD39C")


class TestVariableLockOnVarStatePolicy(unittest.TestCase):
    def test_remaining_buffer(self):
        test_vpl = VariableLockOnVarStatePolicy()
        test_remainder = b'123'
        test_buffer = TEST_GUID_2.bytes_le + b'\x00\x00' + b'\x00A\x00\x00' + test_remainder

        self.assertEqual(test_remainder, test_vpl.decode(test_buffer))

    def test_missing_name(self):
        test_vpl = VariableLockOnVarStatePolicy()

        # Test with no Name field at all.
        test1 = TEST_GUID_1.bytes_le + b'\x00\x00'
        with self.assertRaises(Exception):
            test_vpl.decode(test1)

        # Test with an empty string.
        test2 = test1 + b'\x00\x00'
        with self.assertRaises(Exception):
            test_vpl.decode(test2)

        # Test successful.
        test3 = test1 + b'\x00A\x00\x00'
        _ = test_vpl.decode(test3)

    def test_malformed_name(self):
        test_vpl = VariableLockOnVarStatePolicy()

        # Test with no termination.
        test1 = TEST_GUID_1.bytes_le + b'\x00\x00' + b'\x00A\x00B'
        with self.assertRaises(Exception):
            test_vpl.decode(test1)

        # Test with an unaligned termination.
        test2 = TEST_GUID_1.bytes_le + b'\x00\x00' + b'A\x00B\x00' + b'C' + b'\x00\x00'
        with self.assertRaises(Exception):
            test_vpl.decode(test2)

    def test_to_string(self):
        test_vpl = VariableLockOnVarStatePolicy()
        test_buffer = TEST_GUID_2.bytes_le + b'\x00\x00' + b'A\x00B\x00C\x00\x00\x00'

        test_vpl.decode(test_buffer)

        self.assertEqual(test_vpl.Name, "ABC")


class TestVariablePolicyEntry(unittest.TestCase):
    def test_create_and_to_string(self):
        test_vp = VariablePolicyEntry()
        to_string = str(test_vp)

        # Check for the LockType string.
        self.assertIn("NONE", to_string)

        test_vp.LockPolicyType = VariablePolicyEntry.TYPE_LOCK_ON_CREATE
        to_string = str(test_vp)

        # Check for the new LockType string.
        self.assertIn("CREATE", to_string)

    def test_csv_formatting(self):
        header_row = VariablePolicyEntry.csv_header()
        self.assertIn("Namespace", header_row)
        self.assertIn("LockPolicyType", header_row)

        test_vp = VariablePolicyEntry()
        test_vp.LockPolicyType = VariablePolicyEntry.TYPE_LOCK_ON_CREATE
        csv_row = test_vp.csv_row()
        self.assertEqual(len(header_row), len(csv_row))
        self.assertIn("ON_CREATE", csv_row)

    def test_decoding(self):
        test_vp = VariablePolicyEntry()
        test_vp.decode(TEST_POLICY_ENTRY)

        self.assertEqual(test_vp.Namespace, TEST_POLICY_ENTRY_GUID)
        self.assertEqual(test_vp.LockPolicyType, VariablePolicyEntry.TYPE_LOCK_ON_VAR_STATE)
        self.assertEqual(test_vp.Name, "LastAttemptStatus")
        self.assertEqual(test_vp.LockPolicy.Name, "EOD")

        to_string = str(test_vp)
        self.assertIn("VAR_STATE", to_string)
        self.assertIn("EOD", to_string)
        self.assertIn("LastAttemptStatus", to_string)

    def test_decoding_errors(self):
        test_vp = VariablePolicyEntry()

        with self.assertRaises(ValueError):
            test_vp.decode(TEST_POLICY_ENTRY_BAD_VERSION)
        with self.assertRaises(ValueError):
            test_vp.decode(TEST_POLICY_ENTRY_BAD_LOCK_TYPE)



##
# Utility Functions to support re-use in python scripts.
# Includes functions for running external commands, etc
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import re
import os
import stat
import logging
import datetime
import time
import shutil
import threading
import subprocess
import sys
import inspect
import platform
import importlib
from collections import namedtuple
from enum import Enum as StdEnum
import locale


def Enum(*args):
    items = []
    if len(args) == 1:
        item = args[0]
        if isinstance(item, list):
            items = item
        elif isinstance(item, tuple):
            items = list(item)
        else:
            items.append(item)
    else:
        items = args
    calling_frame = inspect.stack()[1]
    calling_mod = inspect.getmodule(calling_frame[0])
    calling_func = calling_frame.function
    calling_line = calling_frame.lineno

    enum_name = calling_func + ":" + str(calling_line)
    logging.warning(f"Enum is deprecated! Please fix it in {calling_mod}:{calling_func}")
    return StdEnum(enum_name, items, module=calling_mod)


####
# Class to support running commands from the shell in a python environment.
# Don't use directly.
#
# PropagatingThread copied from sample here:
# https://stackoverflow.com/questions/2829329/catch-a-threads-exception-in-the-caller-thread-in-python
####
class PropagatingThread(threading.Thread):
    def run(self):
        self.exc = None
        try:
            if hasattr(self, '_Thread__target'):
                # Thread uses name mangling prior to Python 3.
                self.ret = self._Thread__target(*self._Thread__args, **self._Thread__kwargs)
            else:
                self.ret = self._target(*self._args, **self._kwargs)
        except BaseException as e:
            self.exc = e

    def join(self, timeout=None):
        super(PropagatingThread, self).join()
        if self.exc:
            raise self.exc
        return self.ret


####
# Helper functions for running commands from the shell in python environment
# Don't use directly
#
# process output stream and write to log.
# part of the threading pattern.
#
#  http://stackoverflow.com/questions/19423008/logged-subprocess-communicate
####
def reader(filepath, outstream, stream, logging_level=logging.INFO, encodingErrors='strict'):
    f = None
    # open file if caller provided path
    if(filepath):
        f = open(filepath, "w")

    (_, encoding) = locale.getdefaultlocale()
    while True:
        s = stream.readline().decode(encoding, errors=encodingErrors)
        if not s:
            break
        if(f is not None):
            # write to file if caller provided file
            f.write(s)
        if(outstream is not None):
            # write to stream object if caller provided object
            outstream.write(s)
        logging.log(logging_level, s.rstrip())
    stream.close()
    if(f is not None):
        f.close()


####
# Returns a namedtuple containing information about host machine.
#
# @return namedtuple Host(os=OS Type, arch=System Architecture, bit=Highest Order Bit)
####
def GetHostInfo():
    Host = namedtuple('Host', 'os arch bit')
    host_info = platform.uname()
    os = host_info.system
    processor_info = host_info.machine

    arch = None
    bit = None

    if ("x86" in processor_info.lower()) or ("AMD" in processor_info.upper()) or ("INTEL" in processor_info.upper()):
        arch = "x86"
    elif ("ARM" in processor_info.upper()) or ("AARCH" in processor_info.upper()):
        arch = "ARM"

    if "32" in processor_info:
        bit = "32"
    elif "64" in processor_info:
        bit = "64"

    if (arch is None) or (bit is None):
        raise EnvironmentError("Host info could not be parsed: {0}".format(str(host_info)))

    return Host(os=os, arch=arch, bit=bit)


####
# This is a mixing to do timing on a function. Use it like this:
#
# @timing
# def function_i_want_to_time():
#   ...
####
def timing(f):
    def wrap(*args):
        time1 = time.time()
        ret = f(*args)
        time2 = time.time()
        logging.debug('{:s} function took {:.3f} ms'.format(f.__name__, (time2 - time1) * 1000.0))
        return ret
    return wrap


####
# Run a shell command and print the output to the log file
# This is the public function that should be used to run commands from the shell in python environment
# @param cmd - command being run, either quoted or not quoted
# @param parameters - parameters string taken as is
# @param capture - boolean to determine if caller wants the output captured in any format.
# @param workingdir - path to set to the working directory before running the command.
# @param outfile - capture output to file of given path.
# @param outstream - capture output to a stream.
# @param environ - shell environment variables dictionary that replaces the one inherited from the
#                  current process.
# @param logging_level - log level to log output at.  Default is INFO
# @param raise_exception_on_nonzero - Setting to true causes exception to be raised if the cmd
#                                     return code is not zero.
# @param encodingErrors - may be given to set the desired error handling for encoding errors decoding cmd output.
#                         Default is 'strict'.
#
# @return returncode of called cmd
####
def RunCmd(cmd, parameters, capture=True, workingdir=None, outfile=None, outstream=None, environ=None,
           logging_level=logging.INFO, raise_exception_on_nonzero=False, encodingErrors='strict'):
    cmd = cmd.strip('"\'')
    if " " in cmd:
        cmd = '"' + cmd + '"'
    if parameters is not None:
        parameters = parameters.strip()
        cmd += " " + parameters
    starttime = datetime.datetime.now()
    logging.log(logging_level, "Cmd to run is: " + cmd)
    logging.log(logging_level, "------------------------------------------------")
    logging.log(logging_level, "--------------Cmd Output Starting---------------")
    logging.log(logging_level, "------------------------------------------------")
    c = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=workingdir, shell=True, env=environ)
    if(capture):
        thread = PropagatingThread(target=reader, args=(outfile, outstream, c.stdout, logging_level, encodingErrors))
        thread.start()
        c.wait()
        thread.join()
    else:
        c.wait()

    endtime = datetime.datetime.now()
    delta = endtime - starttime
    endtime_str = "{0[0]:02}:{0[1]:02}".format(divmod(delta.seconds, 60))
    returncode_str = "{0:#010x}".format(c.returncode)
    logging.log(logging_level, "------------------------------------------------")
    logging.log(logging_level, "--------------Cmd Output Finished---------------")
    logging.log(logging_level, "--------- Running Time (mm:ss): " + endtime_str + " ----------")
    logging.log(logging_level, "----------- Return Code: " + returncode_str + " ------------")
    logging.log(logging_level, "------------------------------------------------")

    if raise_exception_on_nonzero and c.returncode != 0:
        raise Exception("{0} failed with Return Code: {1}".format(cmd, returncode_str))
    return c.returncode

####
# Run a python script and print the output to the log file
# This is the public function that should be used to execute python scripts from the shell in python environment.
# The python script will be located using the path as if it was an executable.
#
# @param cmd - cmd string to run including parameters
# @param capture - boolean to determine if caller wants the output captured in any format.
# @param workingdir - path to set to the working directory before running the command.
# @param outfile - capture output to file of given path.
# @param outstream - capture output to a stream.
# @param environ - shell environment variables dictionary that replaces the one inherited from the
#                  current process.
#
# @return returncode of called cmd
####


def RunPythonScript(pythonfile, params, capture=True, workingdir=None, outfile=None, outstream=None,
                    environ=None, logging_level=logging.INFO, raise_exception_on_nonzero=False):
    # locate python file on path
    pythonfile.strip('"\'')
    if " " in pythonfile:
        pythonfile = '"' + pythonfile + '"'
    params.strip()
    logging.debug("RunPythonScript: {0} {1}".format(pythonfile, params))
    if(os.path.isabs(pythonfile)):
        logging.debug("Python Script was given as absolute path: %s" % pythonfile)
    elif(os.path.isfile(os.path.join(os.getcwd(), pythonfile))):
        pythonfile = os.path.join(os.getcwd(), pythonfile)
        logging.debug("Python Script was given as relative path: %s" % pythonfile)
    else:
        # loop thru path environment variable
        for a in os.getenv("PATH").split(os.pathsep):
            a = os.path.normpath(a)
            if os.path.isfile(os.path.join(a, pythonfile)):
                pythonfile = os.path.join(a, pythonfile)
                logging.debug("Python Script was found on the path: %s" % pythonfile)
                break
    params = pythonfile + " " + params
    return RunCmd(sys.executable, params, capture=capture, workingdir=workingdir, outfile=outfile,
                  outstream=outstream, environ=environ, logging_level=logging_level,
                  raise_exception_on_nonzero=raise_exception_on_nonzero)
####
# Locally Sign input file using Windows SDK signtool.  This will use a local Pfx file.
# WARNING!!! : This should not be used for production signing as that process should follow stronger
#               security practices (HSM / smart cards / etc)
#
#  Signing is in format specified by UEFI authentacted variables
####


def DetachedSignWithSignTool(SignToolPath, ToSignFilePath, SignatureOutputFile, PfxFilePath,
                             PfxPass=None, Oid="1.2.840.113549.1.7.2", Eku=None):

    # check signtool path
    if not os.path.exists(SignToolPath):
        logging.error("Path to signtool invalid.  %s" % SignToolPath)
        return -1

    # Adjust for spaces in the path (when calling the command).
    if " " in SignToolPath:
        SignToolPath = '"' + SignToolPath + '"'

    OutputDir = os.path.dirname(SignatureOutputFile)
    # Signtool docs https://docs.microsoft.com/en-us/dotnet/framework/tools/signtool-exe
    # Signtool parameters from
    #   https://docs.microsoft.com/en-us/windows-hardware/manufacture/desktop/secure-boot-key-generation-and-signing-using-hsm--example  # noqa: E501
    # Search for "Secure Boot Key Generation and Signing Using HSM"
    params = 'sign /fd sha256 /p7ce DetachedSignedData /p7co ' + Oid + ' /p7 "' + \
             OutputDir + '" /f "' + PfxFilePath + '"'
    if Eku is not None:
        params += ' /u ' + Eku
    if PfxPass is not None:
        # add password if set
        params = params + ' /p ' + PfxPass
    params = params + ' /debug /v "' + ToSignFilePath + '" '
    ret = RunCmd(SignToolPath, params)
    if(ret != 0):
        logging.error("Signtool failed %d" % ret)
        return ret
    signedfile = os.path.join(OutputDir, os.path.basename(ToSignFilePath) + ".p7")
    if(not os.path.isfile(signedfile)):
        raise Exception("Output file doesn't exist %s" % signedfile)

    shutil.move(signedfile, SignatureOutputFile)
    return ret

####
# Locally Sign input file using Windows SDK signtool.  This will use a local Pfx file.
# WARNING!!! : This should not be used for production signing as that process should follow
#              stronger security practices (HSM / smart cards / etc)
#
#  Signing is catalog format which is an attached signature
####


def CatalogSignWithSignTool(SignToolPath, ToSignFilePath, PfxFilePath, PfxPass=None):

    # check signtool path
    if not os.path.exists(SignToolPath):
        logging.error("Path to signtool invalid.  %s" % SignToolPath)
        return -1

    # Adjust for spaces in the path (when calling the command).
    if " " in SignToolPath:
        SignToolPath = '"' + SignToolPath + '"'

    OutputDir = os.path.dirname(ToSignFilePath)
    # Signtool docs https://docs.microsoft.com/en-us/dotnet/framework/tools/signtool-exe
    # todo: link to catalog signing documentation
    params = "sign /a /fd SHA256 /f " + PfxFilePath
    if PfxPass is not None:
        # add password if set
        params = params + ' /p ' + PfxPass
    params = params + ' /debug /v "' + ToSignFilePath + '" '
    ret = RunCmd(SignToolPath, params, workingdir=OutputDir)
    if(ret != 0):
        logging.error("Signtool failed %d" % ret)
    return ret


###
# Function to print a byte list as hex and optionally output ascii as well as
# offset within the buffer
###
def PrintByteList(ByteList, IncludeAscii=True, IncludeOffset=True, IncludeHexSep=True, OffsetStart=0):
    Ascii = ""
    for index in range(len(ByteList)):
        # Start of New Line
        if(index % 16 == 0):
            if(IncludeOffset):
                print("0x%04X -" % (index + OffsetStart), end='')

        # Midpoint of a Line
        if(index % 16 == 8):
            if(IncludeHexSep):
                print(" -", end='')

        # Print As Hex Byte
        print(" 0x%02X" % ByteList[index], end='')

        # Prepare to Print As Ascii
        if(ByteList[index] < 0x20) or (ByteList[index] > 0x7E):
            Ascii += "."
        else:
            Ascii += ("%c" % ByteList[index])

        # End of Line
        if(index % 16 == 15):
            if(IncludeAscii):
                print(" %s" % Ascii, end='')
            Ascii = ""
            print("")

    # Done - Lets check if we have partial
    if(index % 16 != 15):
        # Lets print any partial line of ascii
        if(IncludeAscii) and (Ascii != ""):
            # Pad out to the correct spot

            while(index % 16 != 15):
                print("     ", end='')
                if(index % 16 == 7):  # account for the - symbol in the hex dump
                    if(IncludeOffset):
                        print("  ", end='')
                index += 1
            # print the ascii partial line
            print(" %s" % Ascii, end='')
            # print a single newline so that next print will be on new line
        print("")


# Simplified Comparison Function borrowed from StackOverflow...
# https://stackoverflow.com/questions/1714027/version-number-comparison
# With Python 3.0 help from:
# https://docs.python.org/3.0/whatsnew/3.0.html#ordering-comparisons
def version_compare(version1, version2):
    def normalize(v):
        return [int(x) for x in re.sub(r'(\.0+)*$', '', v).split(".")]
    (a, b) = (normalize(version1), normalize(version2))
    return (a > b) - (a < b)


def import_module_by_file_name(module_file_path):
    ''' Standard method of importing a Python file. Expecting absolute path. '''
    module_name = os.path.basename(module_file_path)
    spec = importlib.util.spec_from_file_location(module_name, module_file_path)

    if spec is None:
        raise RuntimeError(f"Expected module file named {module_file_path}")

    ImportedModule = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(ImportedModule)

    return ImportedModule


def locate_class_in_module(Module, DesiredClass):
    '''
    Given a module and a class, this function will return the subclass of DesiredClass found in Module.
    It gives preference to classes that are defined in the module itself.
    This means that if you have an import that subclasses DesiredClass, it will be picked unless
    there is a class defined in the module that subclasses DesiredClass.

    In this hypothetical class hierarchy, GrandChildClass would be picked
    --------------      ------------      -----------------
    |DesiredClass|  ->  |ChildClass|  ->  |GrandChildClass|
    --------------      ------------      -----------------
    '''

    DesiredClassInstance = None
    # Pull out the contents of the module that was provided
    module_contents = dir(Module)
    # Filter through the Module, we're only looking for classes.
    classList = [getattr(Module, obj) for obj in module_contents
                 if inspect.isclass(getattr(Module, obj))]

    for _class in classList:
        # Classes that the module import show up in this list too so we need
        # to make sure it's an INSTANCE of DesiredClass, not DesiredClass itself!
        # if multiple instances are found in the same class hierarchy, pick the
        # most specific one. If multiple instances are found belonging to different
        # class hierarchies, raise an error.
        if _class is not DesiredClass and issubclass(_class, DesiredClass):
            if (DesiredClassInstance is None) or issubclass(_class, DesiredClassInstance):
                DesiredClassInstance = _class
            elif not issubclass(DesiredClassInstance, _class):
                raise RuntimeError(f"Multiple instances were found:\n\t{DesiredClassInstance}\n\t{_class}")
    return DesiredClassInstance


def RemoveTree(dir_path: str, ignore_errors: bool = False) -> None:
    '''
    Helper for removing a directory.  Over time there have been
    many private implementations of this due to reliability issues in the
    shutil implementations.  To consolidate on a single function this helper is added.

    On error try to change file attributes.  Also add retry logic.

    dir_path: Path to directory to remove.
    ignore_errors: ignore errors during removal
    '''

    def remove_readonly(func, path, _):
        ''' private function to attempt to change permissions on file/folder being deleted'''
        os.chmod(path, stat.S_IWRITE)
        func(path)

    for _ in range(3):  # retry up to 3 times
        try:
            shutil.rmtree(dir_path, ignore_errors=ignore_errors, onerror=remove_readonly)
        except OSError as err:
            logging.warning(f"Failed to fully remove {dir_path}: {err}")
        else:
            break
    else:
        raise RuntimeError(f"Failed to remove {dir_path}")



## @file
# Module that encodes and decodes a EFI_FIRMWARE_IMAGE_AUTHENTICATION with
# certificate data and payload data.
#
# Copyright (c) 2018 - 2019, Intel Corporation. All rights reserved.<BR>
# Copyright (c) Microsoft Corporation
# SPDX-License-Identifier: BSD-2-Clause-Patent
#

'''
FmpAuthHeader
'''

import struct

from edk2toollib.uefi.wincert import WinCertUefiGuid
from edk2toollib.uefi.edk2.fmp_payload_header import FmpPayloadHeaderClass


class FmpAuthHeaderClass (object):
    # ///
    # /// Image Attribute -Authentication Required
    # ///
    # typedef struct {
    #   ///
    #   /// It is included in the signature of AuthInfo. It is used to ensure freshness/no replay.
    #   /// It is incremented during each firmware image operation.
    #   ///
    #   UINT64                                  MonotonicCount;
    #   ///
    #   /// Provides the authorization for the firmware image operations. It is a signature across
    #   /// the image data and the Monotonic Count value. Caller uses the private key that is
    #   /// associated with a public key that has been provisioned via the key exchange.
    #   /// Because this is defined as a signature, WIN_CERTIFICATE_UEFI_GUID.CertType must
    #   /// be EFI_CERT_TYPE_PKCS7_GUID.
    #   ///
    #   WIN_CERTIFICATE_UEFI_GUID               AuthInfo;
    # } EFI_FIRMWARE_IMAGE_AUTHENTICATION;

    _MonotonicCountFormat = '<Q'
    _MonotonicCountSize = struct.calcsize(_MonotonicCountFormat)

    def __init__(self):
        self.MonotonicCount = 0
        self.AuthInfo = WinCertUefiGuid()
        self.Payload = b''
        self.FmpPayloadHeader = None

    def Encode(self):
        FmpAuthHeader = struct.pack(
            self._MonotonicCountFormat,
            self.MonotonicCount
        )

        if self.FmpPayloadHeader is not None:
            return FmpAuthHeader + self.AuthInfo.Encode() + self.FmpPayloadHeader.Encode()
        else:
            return FmpAuthHeader + self.AuthInfo.Encode() + self.Payload

    def Decode(self, Buffer):
        if len(Buffer) < self._MonotonicCountSize:
            raise ValueError
        (MonotonicCount,) = struct.unpack(
            self._MonotonicCountFormat,
            Buffer[:self._MonotonicCountSize]
        )
        self.MonotonicCount = MonotonicCount

        self.Payload = self.AuthInfo.Decode(Buffer[self._MonotonicCountSize:])
        if len(self.Payload) > 0:
            self.FmpPayloadHeader = FmpPayloadHeaderClass()
            self.FmpPayloadHeader.Decode(self.Payload)
        return self.Payload

    def IsSigned(self, Buffer):
        if len(Buffer) < self._MonotonicCountSize:
            return False

        auth_info = WinCertUefiGuid(Buffer[self._MonotonicCountSize:])
        if auth_info.CertType != WinCertUefiGuid._EFI_CERT_TYPE_PKCS7_GUID.bytes_le:
            return False
        return True

    def DumpInfo(self):
        print('EFI_FIRMWARE_IMAGE_AUTHENTICATION.MonotonicCount                = {MonotonicCount:016X}'
              .format(MonotonicCount=self.MonotonicCount))
        self.AuthInfo.DumpInfo()
        print('sizeof (Payload)                                                = {Size:08X}'
              .format(Size=len(self.Payload)))
        if self.FmpPayloadHeader is not None:
            self.FmpPayloadHeader.DumpInfo()



# @file utility_functions_test.py
# unit test for utility_functions module.
#
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import unittest
import os
import sys
import edk2toollib.utility_functions as utilities


class DesiredClass():
    def __str__(self):
        return "DesiredClass"


class ChildOfDesiredClass(DesiredClass):
    def __str__(self):
        return "Child of DesiredClass"


class GrandChildOfDesiredClass(ChildOfDesiredClass):
    def __str__(self):
        return "GrandChild of DesiredClass"


'''
The current solution can't handle a brother class
class BrotherOfChildOfDesiredClass(DesiredClass):
    def __str__(self):
        return "Brother of Child of DesiredClass"
'''


class UtilityFunctionsTest(unittest.TestCase):

    def test_RunPythonScript(self):  # simple- run yourself!
        path = __file__
        working_dir = os.path.dirname(__file__)
        ret = utilities.RunPythonScript(path, "", working_dir)
        self.assertEqual(ret, 0)
        # test it with all the named parameters
        ret = utilities.RunPythonScript(path, "", capture=False, workingdir=working_dir,
                                        raise_exception_on_nonzero=True)
        self.assertEqual(ret, 0)
        # now try a bad path
        bad_path = __file__ + ".super.bad"
        ret = utilities.RunPythonScript(bad_path, "", capture=False, workingdir=None, raise_exception_on_nonzero=False)
        self.assertNotEqual(ret, 0)
        # now we expect it to throw an exception
        with self.assertRaises(Exception):
            ret = utilities.RunPythonScript(bad_path, "", capture=False, workingdir=None,
                                            raise_exception_on_nonzero=True)
            self.assertNotEqual(ret, 0)

    def test_locate_class_in_module(self):
        module = sys.modules[__name__]

        found_class = utilities.locate_class_in_module(module, DesiredClass)
        self.assertIsNotNone(found_class)
        self.assertEqual(found_class, GrandChildOfDesiredClass)


class EnumClassTest(unittest.TestCase):
    def test_EnumClass_array(self):
        all_animals = ["Dog", "Cat", "Rabbit"]
        animals = utilities.Enum(all_animals)

        # make sure we have three animals
        self.assertEqual(len(animals), 3)
        self.assertTrue(hasattr(animals, "Dog"))
        self.assertFalse(hasattr(animals, "Rat"))
        self.assertFalse(hasattr(animals, "dog"))
        # check to make sure the values are unique
        self.assertNotEqual(animals.Dog, animals.Cat)

        # make sure we can iterate over the members
        for animal in animals.__members__:
            self.assertIn(animal, all_animals)

    def test_EnumClass_args(self):
        colors = utilities.Enum("Green", "Blue", "Red", "Yellow")
        # make sure we have four colors
        self.assertEqual(len(colors), 4)
        self.assertTrue(hasattr(colors, "Red"))
        self.assertFalse(hasattr(colors, "Purple"))
        # check to make sure the values are unique
        self.assertNotEqual(colors.Green, colors.Red)

# DO NOT PUT A MAIN FUNCTION HERE
# this test runs itself to test runpython script, which is a tad bit strange yes.



##
# Module to generate inf files for capsule update. Supports targeting
# multiple ESRT nodes with the same INF.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import textwrap
import re
import uuid
from datetime import datetime
from typing import Optional

SUPPORTED_ARCH = {'amd64': 'amd64',
                  'x64': 'amd64',
                  'arm': 'arm',
                  'arm64': 'ARM64',
                  'aarch64': 'ARM64'
                  }


class InfHeader(object):
    def __init__(self, Name: str, VersionStr: str, CreationDate: str, Arch: str, Provider: str, Manufacturer: str,
                 InfStrings: 'InfStrings') -> None:
        '''Instantiate an INF header object.
        This object represents the INF header at the start of the INF file.

        Name         - specifies the name for the INF package
        VersionStr   - specifies the version as string in dot-tuple format (e.g. "4.15.80.0")
        CreationDate - specifies the INF date as a string in MM/DD/YYYY format (e.g "01/01/2021")
        Arch         - specifies the architecture as a string (e.g. "amd64")
        Provider     - specifies the provider as a string (e.g. "Firmware Provider")
        Manufacturer - Specifies the manufacturer as a string (e.g. "Firmware Manufacturer")
        InfStrings   - An InfStrings object representing the "Strings" section of this INF file.
        '''
        self.Name = Name
        self.VersionStr = VersionStr
        self.Date = CreationDate
        self.Arch = Arch
        InfStrings.AddLocalizableString("Provider", Provider)
        InfStrings.AddLocalizableString("MfgName", Manufacturer)

    @property
    def Name(self):
        return self._name

    @Name.setter
    def Name(self, value):
        # test here for invalid chars
        if not (re.compile(r'[\w-]*$')).match(value):
            raise ValueError("Name has invalid chars.")
        self._name = value

    @property
    def VersionStr(self):
        return self._versionstr

    @VersionStr.setter
    def VersionStr(self, value):
        c = value.count(".")
        if(c < 1) or (c > 3):
            raise ValueError("VersionString must be in format of xx.xx -> xx.xx.xx.xx")
        self._versionstr = value

    @property
    def Date(self):
        return self._date

    @Date.setter
    def Date(self, value):
        self._date = datetime.strptime(value, "%m/%d/%Y").strftime("%m/%d/%Y")

    @property
    def Arch(self):
        return self._arch

    @Arch.setter
    def Arch(self, value):
        key = value.lower()
        if(key not in SUPPORTED_ARCH.keys()):
            raise ValueError("Unsupported Architecture")
        self._arch = SUPPORTED_ARCH[key]

    def __str__(self) -> str:
        '''Return the string representation of this InfHeader object'''
        return textwrap.dedent(f"""\
            ;
            ; {self.Name}
            ; {self.VersionStr}
            ; Copyright (C) Microsoft Corporation. All Rights Reserved.
            ;
            [Version]
            Signature="$WINDOWS NT$"
            Class=Firmware
            ClassGuid={{f2e7dd72-6468-4e36-b6f1-6488f42c1b52}}
            Provider=%Provider%
            DriverVer={self.Date},{self.VersionStr}
            PnpLockdown=1
            CatalogFile={self.Name}.cat

            [Manufacturer]
            %MfgName% = Firmware,NT{self.Arch}

            """)


class InfFirmware(object):
    def __init__(self, Tag: str, Description: str, EsrtGuid: str, VersionInt: str, FirmwareFile: str,
                 InfStrings: 'InfStrings', InfSourceFiles: 'InfSourceFiles', Rollback: Optional[bool] = False,
                 IntegrityFile: Optional[str] = None) -> None:
        '''Instantiate an INF firmware object.
        This object represents individual firmware sections within the INF.

        Tag             - A string that uniquely identifies this firmware (e.g. "Firmware0")
        Description     - A description of the firmware (e.g. "UEFI Firmware")
        EsrtGuid        - ESRT GUID for this firmware in string format. (e.g. "34e094e9-4079-44cd-9450-3f2cb7824c97")
        VersionInt      - Version as an integer in string format (e.g. "1234" or "0x04158000")
        FirmwareFile    - Filename (basename only) of the firmware payload file (e.g. "Firmware1234.bin")
        InfStrings      - An InfStrings object representing the "Strings" section of this INF file.
        InfSourceFiles  - An InfSourceFiles object representing the "SourceDisks*" sections of this INF file.
        Rollback        - Specifies whether this firmware should be enabled for rollback (optional, default: False)
        IntegrityFile   - Filename (basename only) of the integrity file associated with this firmware (e.g.
                          "integrity123.bin"). Optional - if not specified, no integrity file will be included.
        '''
        self.Tag = Tag
        self.Description = Description
        self.EsrtGuid = EsrtGuid
        self.VersionInt = VersionInt
        self.FirmwareFile = FirmwareFile
        InfSourceFiles.AddFile(FirmwareFile)
        self.Rollback = Rollback
        self.IntegrityFile = IntegrityFile
        if (self.IntegrityFile is not None):
            InfSourceFiles.AddFile(IntegrityFile)
        InfStrings.AddNonLocalizableString("REG_DWORD", "0x00010001")

    @property
    def Tag(self):
        return self._tag

    @Tag.setter
    def Tag(self, value):
        # test here for invalid chars
        if not (re.compile(r'[\w-]*$')).match(value):
            raise ValueError("Tag has invalid chars.")
        self._tag = value

    @property
    def EsrtGuid(self):
        return self._esrt_guid

    @EsrtGuid.setter
    def EsrtGuid(self, value):
        uuid.UUID(value)
        self._esrt_guid = value

    @property
    def VersionInt(self):
        return self._version_int

    @VersionInt.setter
    def VersionInt(self, value):
        self._version_int = int(value, base=0)

    def __str__(self) -> str:
        '''Return the string representation of this InfFirmware object'''
        # build rollback string, if required.
        if (self.Rollback):
            RollbackStr = textwrap.dedent(f"""\
                AddReg = {self.Tag}_DowngradePolicy_AddReg

                [{self.Tag}_DowngradePolicy_AddReg]
                HKLM,SYSTEM\\CurrentControlSet\\Control\\FirmwareResources\\{{{self.EsrtGuid}}},Policy,%REG_DWORD%,1
                """)
        else:
            RollbackStr = ""

        # build integrity file string, if required.
        if (self.IntegrityFile is not None):
            IntegrityFile = f"{self.IntegrityFile}\n"
            IntegrityFileReg = f"HKR,,FirmwareIntegrityFilename,,{self.IntegrityFile}\n"
        else:
            IntegrityFile = ""
            IntegrityFileReg = ""

        outstr = textwrap.dedent(f"""\
            [{self.Tag}_Install.NT]
            CopyFiles = {self.Tag}_CopyFiles
            """)
        outstr += RollbackStr
        outstr += textwrap.dedent(f"""
            [{self.Tag}_CopyFiles]
            {self.FirmwareFile}
            """)
        outstr += IntegrityFile
        outstr += textwrap.dedent(f"""
            [{self.Tag}_Install.NT.Hw]
            AddReg = {self.Tag}_AddReg

            [{self.Tag}_AddReg]
            HKR,,FirmwareId,,{{{self.EsrtGuid}}}
            HKR,,FirmwareVersion,%REG_DWORD%,0x{self.VersionInt:X}
            HKR,,FirmwareFilename,,{self.FirmwareFile}
            """)
        outstr += IntegrityFileReg + "\n"
        return outstr


class InfFirmwareSections(object):
    def __init__(self, Arch: str, InfStrings: 'InfStrings') -> None:
        '''Instantiate an INF firmware sections object.
        This object represents a collection of firmware sections and associated common metadata.

        Arch        - specifies the architecture as a string (e.g. "amd64")
        InfStrings  - An InfStrings object representing the "Strings" section of this INF file.
        '''
        self.Arch = Arch
        self.Sections = {}
        self.InfStrings = InfStrings

    @property
    def Arch(self):
        return self._arch

    @Arch.setter
    def Arch(self, value):
        key = value.lower()
        if(key not in SUPPORTED_ARCH.keys()):
            raise ValueError("Unsupported Architecture")
        self._arch = SUPPORTED_ARCH[key]

    def AddSection(self, InfFirmware: InfFirmware) -> None:
        '''Adds an InfFirmware section object to the set of firmware sections in this InfFirmwareSections object.

        InfFirmware - an InfFirmware object representing a firmware section to be added to this collection of sections.
        '''
        self.Sections[InfFirmware.Tag] = InfFirmware
        self.InfStrings.AddLocalizableString(f"{InfFirmware.Tag}Desc", InfFirmware.Description)

    def __str__(self) -> str:
        '''
        Return the string representation of this InfFirmwareSections object (including any InfFirmware objects in it)
        '''
        firmwareStr = f"[Firmware.NT{self.Arch}]\n"
        for InfFirmware in self.Sections.values():
            firmwareStr += f"%{InfFirmware.Tag}Desc% = {InfFirmware.Tag}_Install,UEFI\\RES_{{{InfFirmware.EsrtGuid}}}\n"
        firmwareStr += "\n"
        for InfFirmware in self.Sections.values():
            firmwareStr += str(InfFirmware)
        return firmwareStr


class InfSourceFiles(object):
    def __init__(self, DiskName: str, InfStrings: 'InfStrings') -> None:
        '''Instantiate an INF source files object.
        This object represents the collection of source files that are referenced by other sections of the INF.

        DiskName    - Specifies the DiskName as a string (e.g. "FirmwareUpdate")
        InfStrings  - An InfStrings object representing the "Strings" section of this INF file.
        '''
        self.Files = []
        InfStrings.AddLocalizableString('DiskName', DiskName)
        InfStrings.AddNonLocalizableString('DIRID_WINDOWS', "10")

    def AddFile(self, Filename: str) -> None:
        '''Adds a new file to this InfSourceFiles object

        Filename - Filename (basename only) of the file to be added. (e.g. "Firmware1234.bin")
        '''
        # test here for invalid chars. For filenames, this permits a-z, A-Z, 0-9, _, -, and . characters.
        if not (re.compile(r'[\.\w-]*$')).match(Filename):
            raise ValueError("Filename has invalid chars.")

        if (Filename not in self.Files):
            self.Files.append(Filename)

    def __str__(self) -> str:
        '''Return the string representation of this InfSourceFIles object'''
        Files = ''.join("{0} = 1\n".format(file) for file in self.Files)
        outstr = textwrap.dedent("""\
            [SourceDisksNames]
            1 = %DiskName%

            [SourceDisksFiles]
            """)
        outstr += Files
        outstr += textwrap.dedent("""
            [DestinationDirs]
            DefaultDestDir = %DIRID_WINDOWS%,Firmware ; %SystemRoot%\\Firmware

            """)
        return outstr


class InfStrings(object):
    def __init__(self) -> None:
        '''Instantiate an INF strings object.
        This object represents the collection of strings (localizable or non-localizable) that are referenced by other
        sections of the INF.
        '''
        self.LocalizableStrings = {}
        self.NonLocalizableStrings = {}

    def AddLocalizableString(self, Key: str, Value: str) -> None:
        '''Add a Localizable string to the collection of strings for this INF.

        Key     - the name of this string as it is used in the INF (e.g. "MfgName"). Note: the INF will typically
                  reference this string using % as delimiters (e.g. "%MfgName%"). Do not include the % characters
                  when calling this routine.
        Value   - the value of the localizable string as a string (e.g. "Firmware Manufacturer")
        '''
        # The inf spec says keys here should be "letters, digits, and/or other explicitly visible characters".
        # for simplicity, this just enforces that it must be a string that matches \w+
        if not (re.compile(r'\w+$')).match(Key):
            raise ValueError("Key has invalid chars.")
        self.LocalizableStrings[Key] = Value

    def AddNonLocalizableString(self, Key: str, Value: str) -> None:
        '''Add a Non-Localizable string to the collection of strings for this INF.

        Key     - the name of this string as it is used in the INF (e.g. "REG_DWORD"). Note: the INF will typically
                  reference this string using % as delimiters (e.g. "%REG_DWORD%"). Do not include the % characters
                  when calling this routine.
        Value   - the value of the non-localizable string as a string (e.g. "0x00010001")
        '''
        # The inf spec says keys here should be "letters, digits, and/or other explicitly visible characters".
        # for simplicity, this just enforces that it must be a string that matches \w+
        if not (re.compile(r'\w+$')).match(Key):
            raise ValueError("Key has invalid chars.")
        self.NonLocalizableStrings[Key] = Value

    def __str__(self) -> str:
        '''Return the string representation of this InfStrings object'''
        LocalizedStrings = ""
        LongestKey = max(len(Key) for Key in self.LocalizableStrings.keys())
        for (Key, Value) in self.LocalizableStrings.items():
            LocalizedStrings += '{0:{width}} = "{1}"\n'.format(Key, Value, width=LongestKey)

        NonLocalizedStrings = ""
        LongestKey = max(len(Key) for Key in self.NonLocalizableStrings.keys())
        for (Key, Value) in self.NonLocalizableStrings.items():
            NonLocalizedStrings += "{0:{width}} = {1}\n".format(Key, Value, width=LongestKey)

        outstr = textwrap.dedent("""\
            [Strings]
            ; localizable
            """)
        outstr += LocalizedStrings
        outstr += textwrap.dedent("""
            ; non-localizable
            """)
        outstr += NonLocalizedStrings
        return outstr


class InfFile(object):
    def __init__(self, Name: str, VersionStr: str, CreationDate: str, Provider: str, ManufacturerName: str,
                 Arch: Optional[str] = 'amd64', DiskName: Optional[str] = "Firmware Update") -> None:
        '''Instantiate an INF file object.
        This object represents the entire INF file.

        Users of this implementation are primarily expected to interact with instances of this class.

        Name         - specifies the name for the INF package
        VersionStr   - specifies the version as string in dot-tuple format (e.g. "4.15.80.0")
        CreationDate - specifies the INF date as a string in MM/DD/YYYY format (e.g "01/01/2021")
        Provider     - specifies the provider as a string (e.g. "Firmware Provider")
        Manufacturer - Specifies the manufacturer as a string (e.g. "Firmware Manufacturer")
        Arch         - specifies the architecture as a string. Optional, defaults to "amd64".
        DiskName     - specifies the "Disk Name" for this update. Optional, defaults to "Firmware Update".
        '''
        self.InfStrings = InfStrings()
        self.InfSourceFiles = InfSourceFiles(DiskName, self.InfStrings)
        self.InfHeader = InfHeader(Name, VersionStr, CreationDate, Arch, Provider, ManufacturerName, self.InfStrings)
        self.InfFirmwareSections = InfFirmwareSections(Arch, self.InfStrings)

    def AddFirmware(self, Tag: str, Description: str, EsrtGuid: str, VersionInt: str, FirmwareFile: str,
                    Rollback: Optional[bool] = False, IntegrityFile: Optional[str] = None) -> None:
        '''Adds a firmware target to the INF.

        Tag             - A string that uniquely identifies this firmware (e.g. "Firmware0")
        Description     - A description of the firmware (e.g. "UEFI Firmware")
        EsrtGuid        - ESRT GUID for this firmware in string format. (e.g. "34e094e9-4079-44cd-9450-3f2cb7824c97")
        VersionInt      - Version as an integer in string format (e.g. "1234" or "0x04158000")
        FirmwareFile    - Filename (basename only) of the firmware payload file (e.g. "Firmware1234.bin")
        Rollback        - Specifies whether this firmware should be enabled for rollback (optional, default: False)
        IntegrityFile   - Filename (basename only) of the integrity file associated with this firmware (e.g.
                          "integrity123.bin"). Optional - if not specified, no integrity file will be included.
        '''
        firmwareSection = InfFirmware(Tag, Description, EsrtGuid, VersionInt, FirmwareFile,
                                      self.InfStrings, self.InfSourceFiles, Rollback, IntegrityFile)
        self.InfFirmwareSections.AddSection(firmwareSection)

    def __str__(self) -> str:
        '''
        Returns the string representation of this InfFile object. The resulting string is suitable for writing to an
        INF file for inclusion in a capsule package.'''
        return str(self.InfHeader) + str(self.InfFirmwareSections) + str(self.InfSourceFiles) + str(self.InfStrings)



# @file
# Module contains defintions and structures from the UefiMultiPhase header file.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import struct

EFI_VARIABLE_NON_VOLATILE = 0x00000001
EFI_VARIABLE_BOOTSERVICE_ACCESS = 0x00000002
EFI_VARIABLE_RUNTIME_ACCESS = 0x00000004
EFI_VARIABLE_HARDWARE_ERROR_RECORD = 0x00000008
EFI_VARIABLE_AUTHENTICATED_WRITE_ACCESS = 0x00000010
EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS = 0x00000020
EFI_VARIABLE_APPEND_WRITE = 0x00000040


class EfiVariableAttributes(object):
    # UINT32
    _StructFormat = "<I"
    _StructSize = struct.calcsize(_StructFormat)

    STRING_MAP = {
        EFI_VARIABLE_APPEND_WRITE: "EFI_VARIABLE_APPEND_WRITE",
        EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS: "EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS",
        EFI_VARIABLE_AUTHENTICATED_WRITE_ACCESS: "EFI_VARIABLE_AUTHENTICATED_WRITE_ACCESS",
        EFI_VARIABLE_HARDWARE_ERROR_RECORD: "EFI_VARIABLE_HARDWARE_ERROR_RECORD",
        EFI_VARIABLE_RUNTIME_ACCESS: "EFI_VARIABLE_RUNTIME_ACCESS",
        EFI_VARIABLE_BOOTSERVICE_ACCESS: "EFI_VARIABLE_BOOTSERVICE_ACCESS",
        EFI_VARIABLE_NON_VOLATILE: "EFI_VARIABLE_NON_VOLATILE",
    }

    def __init__(self, attributes=0x0000_0000):
        self.Attributes = attributes

    def __str__(self):
        result = []
        for key in EfiVariableAttributes.STRING_MAP:
            if self.Attributes & key:
                result.append(EfiVariableAttributes.STRING_MAP[key])
        return ",".join(result)



# @file string_handler_test.py
# Contains unit test routines for the string_handler functions
#
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import unittest
import logging
from edk2toollib.log.string_handler import StringStreamHandler


class TestStringStreamHandler(unittest.TestCase):

    def test_init(self):
        handler = StringStreamHandler()
        self.assertNotEqual(handler, None)

    def create_record(self, message="TEST", level=logging.INFO, name=""):
        return logging.LogRecord(name, level, __file__, 0, message, [], None)

    def test_readlines(self):
        # create the handler
        handler = StringStreamHandler()
        handler.setLevel(logging.DEBUG)
        LINES_TO_DEBUG = 10
        # create some records for it to process
        for i in range(LINES_TO_DEBUG):
            rec = self.create_record(f"test{i}")
            handler.handle(rec)
        # check to make sure we don't have any
        self.assertEqual(len(handler.readlines()), 0, "We shouldn't have anything because our stream is at the end")
        # go to the beginning and read the streams
        handler.seek_start()
        self.assertEqual(len(handler.readlines()), LINES_TO_DEBUG, "We should have at least a few")
        # go to the beginning but then back to the end
        handler.seek_start()
        handler.seek_end()
        self.assertEqual(len(handler.readlines()), 0, "We shouldn't have anything because our stream is at the end")



# @file dsc_parser_test.py
# Contains unit test routines for the dsc parser class.
#
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import unittest
import tempfile
import os
from edk2toollib.uefi.edk2.parsers.dsc_parser import DscParser


class TestDscParserBasic(unittest.TestCase):

    def test_creation(self):
        a = DscParser()
        self.assertNotEqual(a, None)


class TestDscParserIncludes(unittest.TestCase):

    @staticmethod
    def write_to_file(file_path, data):
        f = open(file_path, "w")
        f.writelines(data)
        f.close()

    def test_dsc_include_single_file(self):
        ''' This tests whether includes work properly '''
        workspace = tempfile.mkdtemp()

        file1_name = "file1.dsc"
        file2_name = "file2.dsc"
        file1_path = os.path.join(workspace, file1_name)
        file2_path = os.path.join(workspace, file2_name)

        file1_data = f"!include {file2_name}"
        file2_data = "[Defines]\nINCLUDED = TRUE"

        TestDscParserIncludes.write_to_file(file1_path, file1_data)
        TestDscParserIncludes.write_to_file(file2_path, file2_data)

        parser = DscParser()
        parser.SetBaseAbsPath(workspace)
        parser.ParseFile(file1_path)

        # test to make sure we did it right
        self.assertEqual(len(parser.LocalVars), 1)  # make sure we only have one define
        self.assertEqual(parser.LocalVars["INCLUDED"], "TRUE")  # make sure we got the defines
        self.assertEqual(len(parser.GetAllDscPaths()), 2)  # make sure we have two dsc paths

    def test_dsc_include_missing_file(self):
        ''' This tests whether includes work properly '''
        workspace = tempfile.mkdtemp()

        file1_name = "file1.dsc"
        file1_path = os.path.join(workspace, file1_name)

        file1_data = "!include BAD_FILE.dsc"

        TestDscParserIncludes.write_to_file(file1_path, file1_data)

        parser = DscParser()
        parser.SetBaseAbsPath(workspace)
        with self.assertRaises(FileNotFoundError):
            parser.ParseFile(file1_path)

    def test_dsc_include_missing_file_no_fail_mode(self):
        ''' This tests whether includes work properly if no fail mode is on'''
        workspace = tempfile.mkdtemp()

        file1_name = "file1.dsc"
        file1_path = os.path.join(workspace, file1_name)

        file1_data = "!include BAD_FILE.dsc"

        TestDscParserIncludes.write_to_file(file1_path, file1_data)

        parser = DscParser()
        parser.SetNoFailMode()
        parser.SetBaseAbsPath(workspace)
        parser.ParseFile(file1_path)

    def test_dsc_parse_file_on_package_path(self):
        ''' This tests whether includes work properly if no fail mode is on'''
        workspace = tempfile.mkdtemp()
        working_dir_name = "working"
        working2_dir_name = "working2"

        working_folder = os.path.join(workspace, working_dir_name)
        working2_folder = os.path.join(working_folder, working2_dir_name)
        os.makedirs(working_folder, exist_ok=True)
        os.makedirs(working2_folder, exist_ok=True)

        file1_name = "file1.dsc"
        file1_path = os.path.join(working2_folder, file1_name)
        file1_short_path = os.path.join(working2_dir_name, file1_name)
        file1_data = "[Defines]\n INCLUDED=TRUE"

        TestDscParserIncludes.write_to_file(file1_path, file1_data)
        with self.assertRaises(FileNotFoundError):
            parser = DscParser()
            parser.SetBaseAbsPath(workspace)
            parser.ParseFile(file1_short_path)

        parser = DscParser()
        parser.SetBaseAbsPath(workspace)
        parser.SetPackagePaths([working_folder, ])
        parser.ParseFile(file1_short_path)
        self.assertEqual(parser.LocalVars["INCLUDED"], "TRUE")  # make sure we got the defines

    def test_dsc_include_relative_path(self):
        ''' This tests whether includes work properly with a relative path'''
        workspace = tempfile.mkdtemp()
        outside_folder = os.path.join(workspace, "outside")
        inside_folder = os.path.join(outside_folder, "inside")
        inside2_folder = os.path.join(outside_folder, "inside2")
        random_folder = os.path.join(outside_folder, "random")
        os.makedirs(inside_folder, exist_ok=True)
        os.makedirs(inside2_folder, exist_ok=True)
        os.makedirs(random_folder, exist_ok=True)
        cwd = os.getcwd()
        os.chdir(random_folder)
        try:

            file1_name = "file1.dsc"
            file1_path = os.path.join(outside_folder, file1_name)

            file2_name = "file2.dsc"
            file2_path = os.path.join(inside_folder, file2_name)

            file3_name = "file3.dsc"
            file3_path = os.path.join(inside2_folder, file3_name)

            file1_data = "!include " + os.path.relpath(file2_path, os.path.dirname(file1_path)).replace("\\", "/")
            file2_data = "!include " + os.path.relpath(file3_path, os.path.dirname(file2_path)).replace("\\", "/")
            file3_data = "[Defines]\n INCLUDED=TRUE"

            print(f"{file1_path}: {file1_data}")
            print(f"{file2_path}: {file2_data}")
            print(f"{file3_path}: {file3_data}")

            TestDscParserIncludes.write_to_file(file1_path, file1_data)
            TestDscParserIncludes.write_to_file(file2_path, file2_data)
            TestDscParserIncludes.write_to_file(file3_path, file3_data)

            parser = DscParser()
            parser.SetBaseAbsPath(workspace)
            parser.ParseFile(file1_path)

            self.assertEqual(parser.LocalVars["INCLUDED"], "TRUE")  # make sure we got the defines
        finally:
            os.chdir(cwd)



# @file BaseParser.py
# Code to support parsing EDK2 files
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import os
import logging
from edk2toollib.uefi.edk2 import path_utilities


class BaseParser(object):
    """ """
    operators = ["OR", "AND", "IN", "==", "!=", ">", "<", "<=", ">="]

    def __init__(self, log="BaseParser"):
        self.Logger = logging.getLogger(log)
        self.Lines = []
        self.LocalVars = {}
        self.InputVars = {}
        self.CurrentSection = ""
        self.CurrentFullSection = ""
        self.Parsed = False
        self.ConditionalStack = []
        self.RootPath = ""
        self.PPs = []
        self._Edk2PathUtil = None
        self.TargetFilePath = None  # the abs path of the target file
        self.CurrentLine = -1
        self._MacroNotDefinedValue = "0"  # value to used for undefined macro

    #
    # For include files set the base root path
    #

    def SetBaseAbsPath(self, path):
        """

        Args:
          path:

        Returns:

        """
        self.RootPath = os.path.abspath(path)
        self._ConfigEdk2PathUtil()
        return self

    def _ConfigEdk2PathUtil(self):
        ''' creates the path utility object based on the root path and package paths '''
        self._Edk2PathUtil = path_utilities.Edk2Path(self.RootPath, self.PPs, error_on_invalid_pp=False)

    def SetPackagePaths(self, pps=[]):
        """

        Args:
          pps:  (Default value = [])

        This must be called after SetBaseAbsPath

        Returns:

        """
        self.PPs = pps
        self._ConfigEdk2PathUtil()
        return self

    def SetInputVars(self, inputdict):
        """

        Args:
          inputdict:

        Returns:

        """
        self.InputVars = inputdict
        return self

    def FindPath(self, *p):
        """
        Given a path, it will find it relative to the root, the current target file, or the packages path
        Args:
          *p: any number of strings or path like objects

        Returns: a full absolute path if the file exists, None on failure

        """
        # check if we're getting a None
        if p is None or (len(p) == 1 and p[0] is None):
            return None

        Path = os.path.join(*p)
        # check if it it is absolute
        if os.path.isabs(Path) and os.path.exists(Path):
            return Path

        # If the absolute path exists, return it.
        Path = os.path.join(self.RootPath, *p)
        if os.path.exists(Path):
            return os.path.abspath(Path)

        # If that fails, check a path relative to the target file.
        if self.TargetFilePath is not None:
            Path = os.path.abspath(os.path.join(os.path.dirname(self.TargetFilePath), *p))
            if os.path.exists(Path):
                return os.path.abspath(Path)

        # If that fails, check in every possible Pkg path.
        if self._Edk2PathUtil is not None:
            target_path = os.path.join(*p)
            Path = self._Edk2PathUtil.GetAbsolutePathOnThisSystemFromEdk2RelativePath(target_path, False)
            if Path is not None:
                return Path

        # log invalid file path
        self.Logger.error(f"Invalid file path: {p}")
        return None

    def WriteLinesToFile(self, filepath):
        """

        Args:
          filepath:

        Returns:

        """
        self.Logger.debug("Writing all lines to file: %s" % filepath)
        file_handle = open(filepath, "w")
        for line in self.Lines:
            file_handle.write(line + "\n")
        file_handle.close()
    #
    # do logical comparisons
    #

    def ComputeResult(self, value, cond, value2):
        """

        Args:
          value:
          cond:
          value2:

        Returns:

        """
        ivalue = value
        ivalue2 = value2
        if isinstance(value, str):
            ivalue = value.strip("\"")
        if isinstance(value2, str):
            ivalue2 = value2.strip("\"")

        # convert it to interpretted value
        if (cond.upper() == "IN"):
            # strip quotes
            self.Logger.debug(f"{ivalue} in {ivalue2}")

            return ivalue in ivalue2

        try:
            ivalue = self.ConvertToInt(ivalue)
        except ValueError:
            pass
        try:
            if(cond.lower() == "in"):
                ivalue2 = set(ivalue2.split())
            else:
                ivalue2 = self.ConvertToInt(ivalue2)
        except ValueError:
            pass

        # First check our boolean operators
        if (cond.upper() == "OR"):
            return ivalue or ivalue2
        if (cond.upper() == "AND"):
            return ivalue and ivalue2

        # check our truthyness
        if(cond == "=="):
            # equal
            return (ivalue == ivalue2) or (value == value2)

        elif (cond == "!="):
            # not equal
            return (ivalue != ivalue2) and (value != value2)

        # check to make sure we only have digits from here on out
        if not isinstance(value, int) and not str.isdigit(value):
            self.Logger.error(f"{self.__class__}: Unknown value: {value} {ivalue.__class__}")
            self.Logger.debug(f"{self.__class__}: Conditional: {value} {cond}{value2}")
            raise ValueError("Unknown value")

        if not isinstance(value2, int) and not str.isdigit(value2):
            self.Logger.error(f"{self.__class__}: Unknown value: {value2} {ivalue2}")
            self.Logger.debug(f"{self.__class__}: Conditional: {value} {cond} {value2}")
            raise ValueError("Unknown value")

        if (cond == "<"):
            return (ivalue < ivalue2)

        elif (cond == "<="):
            return (ivalue <= ivalue2)

        elif (cond == ">"):
            return (ivalue > ivalue2)

        elif (cond == ">="):
            return (ivalue >= ivalue2)

        else:
            self.Logger.error(f"{self.__class__}: Unknown conditional: {cond}")
            raise RuntimeError("Unknown conditional")

    #
    # convert to int based on prefix
    #

    def ConvertToInt(self, value):
        """

        Args:
          value: must be str or int

        Returns:

        """
        if isinstance(value, int):
            return value
        if isinstance(value, str) and value.upper() == "TRUE":
            return 1
        elif isinstance(value, str) and value.upper() == "FALSE":
            return 0
        elif isinstance(value, str) and value.upper().startswith("0X"):
            return int(value, 16)
        else:
            return int(value, 10)

    #
    # Push new value on stack
    #
    def PushConditional(self, v):
        """

        Args:
          v:

        Returns:

        """
        self.ConditionalStack.append(v)

    #
    # Pop conditional and return the value
    #

    def PopConditional(self):
        """ """
        if(len(self.ConditionalStack) > 0):
            return self.ConditionalStack.pop()
        else:
            self.Logger.critical("Tried to pop an empty conditional stack.  Line Number %d" % self.CurrentLine)
            return self.ConditionalStack.pop()  # this should cause a crash but will give trace.

    def _FindReplacementForToken(self, token, replace_if_not_found=False):

        v = self.LocalVars.get(token)

        if(v is None):
            v = self.InputVars.get(token)

        if(v is None and replace_if_not_found):
            v = self._MacroNotDefinedValue

        elif(v is None):
            return None

        if (type(v) is bool):
            v = "true" if v else "false"

        if(type(v) is str and (v.upper() == "TRUE" or v.upper() == "FALSE")):
            v = v.upper()

        return str(v)

    #
    # Method to replace variables
    # in a line with their value from input dict or local dict
    #
    def ReplaceVariables(self, line):
        """

        Args:
          line:

        Returns:

        """
        # first tokenize and look for tokens require special macro
        # handling without $.  This must be done first otherwise
        # both syntax options can not be supported.
        result = line
        tokens = result.split()
        replace = len(tokens) > 1 and tokens[0].lower() in ["!ifdef", "!ifndef", "!if", "!elseif"]
        if len(tokens) > 1 and tokens[0].lower() in ["!ifdef", "!ifndef"]:
            if not tokens[1].startswith("$("):
                v = self._FindReplacementForToken(tokens[1], replace)
                if v is not None:
                    result = result.replace(tokens[1], v, 1)

        # use line to avoid change by handling above
        rep = line.count("$")
        index = 0
        while(rep > 0):
            start = line.find("$(", index)
            end = line.find(")", start)

            token = line[start + 2:end]
            replacement_token = line[start:end + 1]
            self.Logger.debug("Token is %s" % token)
            v = self._FindReplacementForToken(token, replace)
            if v is not None:
                result = result.replace(replacement_token, v, 1)

            index = end + 1
            rep = rep - 1

        return result

    #
    # Process Conditional
    # return true if line is a conditional otherwise false
    #

    def ProcessConditional(self, text):
        """

        Args:
          text:

        Returns:

        """
        if '"' in text:
            tokens = text.split('"')
            tokens = tokens[0].split() + [tokens[1]] + tokens[2].split()
        else:
            tokens = text.split()
        if(tokens[0].lower() == "!if"):
            self.PushConditional(self.EvaluateConditional(text))
            return True

        elif(tokens[0].lower() == "!ifdef"):
            if len(tokens) != 2:
                self.Logger.error("!ifdef conditionals need to be formatted correctly (spaces between each token)")
                raise RuntimeError("Invalid conditional", text)
            self.PushConditional((tokens[1] != self._MacroNotDefinedValue))
            return True

        elif(tokens[0].lower() == "!ifndef"):
            if len(tokens) != 2:
                self.Logger.error("!ifdef conditionals need to be formatted correctly (spaces between each token)")
                raise RuntimeError("Invalid conditional", text)
            self.PushConditional((tokens[1] == self._MacroNotDefinedValue))
            return True

        elif(tokens[0].lower() == "!else"):
            if len(tokens) != 1:
                self.Logger.error("!ifdef conditionals need to be formatted correctly (spaces between each token)")
                raise RuntimeError("Invalid conditional", text)
            v = self.PopConditional()
            # TODO make sure we can't do multiple else statements
            self.PushConditional(not v)
            return True

        elif(tokens[0].lower() == "!endif"):
            if len(tokens) != 1:
                self.Logger.error("!ifdef conditionals need to be formatted correctly (spaces between each token)")
                raise RuntimeError("Invalid conditional", text)
            self.PopConditional()
            return True

        return False

    def EvaluateConditional(self, text):
        ''' Uses a pushdown resolver '''
        text = str(text).strip()
        if not text.lower().startswith("!if "):
            raise RuntimeError(f"Invalid conditional cannot be validated: {text}")
        text = text[3:].strip()
        self.Logger.debug(f"STAGE 1: {text}")
        text = self.ReplaceVariables(text)
        self.Logger.debug(f"STAGE 2: {text}")
        tokens = self._TokenizeConditional(text)
        self.Logger.debug(f"STAGE 3: {tokens}")
        expression = self._ConvertTokensToPostFix(tokens)
        self.Logger.debug(f"STAGE 4: {expression}")

        # Now we evaluate the post fix expression
        if len(expression) == 0:
            raise RuntimeError(f"Malformed !if conditional expression {text} {expression}")
        while len(expression) != 1:
            first_operand_index = -1
            # find the first operator
            for index, item in enumerate(expression):
                if self._IsOperator(item):
                    first_operand_index = index
                    break
            if first_operand_index == -1:
                raise RuntimeError(f"We didn't find an operator to execute in {expression}: {text}")
            operand = expression[first_operand_index]

            if operand == "NOT":
                # Special logic for handling the not
                if first_operand_index < 1:
                    raise RuntimeError(f"We have a stray operand {operand}")
                # grab the operand right before the NOT and invert it
                operator1_raw = expression[first_operand_index - 1]
                operator1 = self.ConvertToInt(operator1_raw)
                result = not operator1
                # grab what was before the operator and the operand, then squish it all together
                new_expression = expression[:first_operand_index - 1] if first_operand_index > 1 else []
                new_expression += [result, ] + expression[first_operand_index + 1:]
                expression = new_expression
            else:
                if first_operand_index < 2:
                    raise RuntimeError(f"We have a stray operand {operand}")
                operator1 = expression[first_operand_index - 2]
                operator2 = expression[first_operand_index - 1]

                do_invert = False
                # check if we have a special operator that has a combined not on it
                if str(operand).startswith("!+"):
                    operand = operand[2:]
                    do_invert = True
                # compute the result now that we have the three things we need
                result = self.ComputeResult(operator1, operand, operator2)

                if do_invert:
                    result = not result
                # grab what was before the operator and the operand, then smoosh it all together
                new_expression = expression[:first_operand_index - 2] if first_operand_index > 2 else []
                new_expression += [result, ] + expression[first_operand_index + 1:]
                expression = new_expression

        final = self.ConvertToInt(expression[0])
        self.Logger.debug(f" FINAL {expression} {final}")

        return bool(final)

    @classmethod
    def _TokenizeConditional(cls, text):
        ''' takes in a string that has macros replaced '''
        # TOKENIZER
        # first we create tokens
        TEXT_MODE = 0
        QUOTE_MODE = 1
        MACRO_MODE = 2
        token = ""
        mode = 0
        tokens = []
        for character in text:
            if character == "\"" and len(token) == 0:
                mode = QUOTE_MODE
            elif character == "\"" and mode == QUOTE_MODE:
                if len(token) > 0:
                    tokens.append(f"\"{token}\"")
                    token = ""
                mode = TEXT_MODE
            elif character == "$" and len(token) == 0:
                token += character
                mode = MACRO_MODE
            elif character == ')' and mode == MACRO_MODE:
                token += character
                tokens.append(token)
                token = ""
                mode = TEXT_MODE
            elif mode == TEXT_MODE and (character == "(" or character == ")"):
                if len(token) > 0:
                    tokens.append(token)
                    token = ""
                tokens.append(character)
            elif character == " " and (mode == TEXT_MODE or mode == MACRO_MODE):
                if len(token) > 0:
                    tokens.append(token)
                    token = ""
                    mode = TEXT_MODE
            else:
                token += character
        # make sure to add in the last token just in case
        if len(token) > 0:
            tokens.append(token)

        # then we do the lexer and convert operands as necessary
        for index in range(len(tokens)):
            token = tokens[index]
            token_upper = token.upper()
            if token_upper in cls.operators:
                token = token_upper
            elif token_upper == "||":
                token = "OR"
            elif token_upper == "&&":
                token = "AND"
            elif token_upper == "EQ":
                token = "=="
            elif token_upper == "NE":
                token = "!="
            elif token == "!":
                token = "NOT"
            tokens[index] = token

        # collapse the not
        collapsed_tokens = []
        found_not = False
        for token in tokens:
            if str(token).upper() == "NOT":
                found_not = True
                continue
            if not found_not:
                collapsed_tokens.append(token)
            elif token in cls.operators:
                collapsed_tokens.append("!+" + token)
                found_not = False
            else:  # add the not back
                found_not = False
                collapsed_tokens.append("NOT")
                collapsed_tokens.append(token)

        return collapsed_tokens

    @classmethod
    def _ConvertTokensToPostFix(cls, tokens):
        # convert infix into post fix
        stack = ["("]
        tokens.append(")")  # add an extra parathesis
        expression = []
        for token in tokens:
            # If the incoming symbol is a left parenthesis, push it on the stack.
            if token == "(":
                stack.append(token)
            # If the incoming symbol is a right parenthesis,
            # pop the stack and print the operators until you see a left parenthesis.
            # Discard the pair of parentheses.
            elif token == ")":
                while len(stack) > 0 and stack[-1] != '(':
                    expression.append(stack.pop())
                stack.pop()  # pop the last (
            # If this isn't a operator ignore it
            elif not cls._IsOperator(token):
                expression.append(token)
            # If the stack is empty or contains a left parenthesis on top, push the incoming operator onto the stack.
            elif len(stack) == 0 or stack[-1] == '(':
                stack.append(token)
            # If the incoming symbol has higher precedence than the top of the stack, push it on the stack.
            elif len(stack) == 0 or cls._GetOperatorPrecedence(token) > cls._GetOperatorPrecedence(stack[-1]):
                stack.append(token)
            # If the incoming symbol has equal precedence with the top of the stack, use association.
            # If the association is left to right, pop and print the top of the stack and
            # then push the incoming operator. If the association is right to left, push the incoming operator.
            elif len(stack) != 0 and cls._GetOperatorPrecedence(token) == cls._GetOperatorPrecedence(stack[-1]):
                expression.append(stack.pop())
                stack.append(token)
            # If the incoming symbol has lower precedence than the symbol on the top of the stack,
            # pop the stack and print the top operator.
            # Then test the incoming operator against the new top of stack.
            elif len(stack) != 0 and cls._GetOperatorPrecedence(token) < cls._GetOperatorPrecedence(stack[-1]):
                while len(stack) > 0 and cls._GetOperatorPrecedence(token) <= cls._GetOperatorPrecedence(stack[-1]):
                    expression.append(stack.pop())
                stack.append(token)
            else:
                logging.error("We don't know what to do with " + token)
        while len(stack) > 0:
            val = stack.pop()
            expression.append(val)
        return expression

    @classmethod
    def _IsOperator(cls, token):
        if type(token) is not str:
            return False
        if token.startswith("!+"):
            token = token[2:]
        if token == "NOT":  # technically an operator
            return True
        return token in cls.operators

    @classmethod
    def _GetOperatorPrecedence(cls, token):
        if not cls._IsOperator(token):
            return -1
        if token == "(" or token == ")":
            return 100
        if token == "NOT":  # not is the lowest
            return -2
        if token == "IN":
            return 1
        return 0

    #
    # returns true or false depending on what state of conditional you are currently in
    #

    def InActiveCode(self):
        """ """
        ret = True
        for a in self.ConditionalStack:
            if not a:
                ret = False
                break

        return ret

    def IsGuidString(self, line):
        """
        will return true if the the line has
        = { 0xD3B36F2C, 0xD551, 0x11D4, { 0x9A, 0x46, 0x00, 0x90, 0x27, 0x3F, 0xC1, 0x4D }}
        Args:
          line:

        Returns:

        """
        if(line.count("{") == 2 and line.count("}") == 2 and line.count(",") == 10 and line.count("=") == 1):
            return True
        return False

    def ParseGuid(self, line):
        """
        parse a guid into a different format
        Will throw exception if missing any of the 11 parts of isn't long enough
        Args:
          l: the guid to parse ex: { 0xD3B36F2C, 0xD551, 0x11D4, { 0x9A, 0x46, 0x00, 0x90, 0x27, 0x3F, 0xC1, 0x4D }}

        Returns: a string of the guid. ex: D3B36F2C-D551-11D4-9A46-0090273FC14D

        """
        entries = line.lstrip(' {').rstrip(' }').split(',')
        if len(entries) != 11:
            raise RuntimeError(
                f"Invalid GUID found {line}. We are missing some parts since we only found: {len(entries)}")
        gu = entries[0].lstrip(' 0').lstrip('x').strip()
        # pad front until 8 chars
        while(len(gu) < 8):
            gu = "0" + gu

        gut = entries[1].lstrip(' 0').lstrip('x').strip()
        while(len(gut) < 4):
            gut = "0" + gut
        gu = gu + "-" + gut

        gut = entries[2].lstrip(' 0').lstrip('x').strip()
        while(len(gut) < 4):
            gut = "0" + gut
        gu = gu + "-" + gut

        # strip off extra {
        gut = entries[3].lstrip(' { 0').lstrip('x').strip()
        while(len(gut) < 2):
            gut = "0" + gut
        gu = gu + "-" + gut

        gut = entries[4].lstrip(' 0').lstrip('x').strip()
        while(len(gut) < 2):
            gut = "0" + gut
        gu = gu + gut

        gut = entries[5].lstrip(' 0').lstrip('x').strip()
        while(len(gut) < 2):
            gut = "0" + gut
        gu = gu + "-" + gut

        gut = entries[6].lstrip(' 0').lstrip('x').strip()
        while(len(gut) < 2):
            gut = "0" + gut
        gu = gu + gut

        gut = entries[7].lstrip(' 0').lstrip('x').strip()
        while(len(gut) < 2):
            gut = "0" + gut
        gu = gu + gut

        gut = entries[8].lstrip(' 0').lstrip('x').strip()
        while(len(gut) < 2):
            gut = "0" + gut
        gu = gu + gut

        gut = entries[9].lstrip(' 0').lstrip('x').strip()
        while(len(gut) < 2):
            gut = "0" + gut
        gu = gu + gut

        gut = entries[10].split()[0].lstrip(' 0').lstrip('x').rstrip(' } ').strip()
        while(len(gut) < 2):
            gut = "0" + gut
        gu = gu + gut

        proper_guid_length = 36
        if len(gu) > proper_guid_length:
            raise RuntimeError(f"The guid we parsed was too long: {gu}")
        if len(gu) < proper_guid_length:
            raise RuntimeError(f"The guid we parsed was too short: {gu}")

        return gu.upper()

    def ResetParserState(self):
        """ """
        self.ConditionalStack = []
        self.CurrentSection = ''
        self.CurrentFullSection = ''
        self.Parsed = False

#
# Base Class for Edk2 build files that use # for comments
#


class HashFileParser(BaseParser):
    """ """

    def __init__(self, log):
        BaseParser.__init__(self, log)

    def StripComment(self, line):
        """

        Args:
          line:

        Returns:

        """
        return line.split('#')[0].strip()

    def ParseNewSection(self, line):
        """

        Args:
          line:

        Returns:

        """
        if(line.count("[") == 1 and line.count("]") == 1):  # new section
            section = line.strip().lstrip("[").split(".")[0].split(",")[0].rstrip("]").strip()
            self.CurrentFullSection = line.strip().lstrip("[").split(",")[0].rstrip("]").strip()
            return (True, section)
        return (False, "")



##
# Copyright (C) Microsoft Corporation. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause-Patent
#
# Python script that converts a raw IVRS table into a struct
##
# spell-checker:ignore IVMD, IOMMUEFR

import unittest
from edk2toollib.acpi.dmar_parser import DMARTable


class DmarParserTest(unittest.TestCase):

    dse_t_01h = None
    dse_t_03h = None
    dse_t_04h = None

    drhd_t = None
    rmrr_t = None

    def test_dmar_parser_dse_01h(self):
        # PCI Endpoint Device
        dse_t_01h = bytes([0x01, 0x08, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00])
        DmarParserTest.dse_t_01h = DMARTable.DeviceScopeStruct(dse_t_01h)
        self.assertNotEqual(DmarParserTest.dse_t_01h, None)
        self.assertEqual(DmarParserTest.dse_t_01h.Type, 0x01)
        self.assertEqual(DmarParserTest.dse_t_01h.Length, 0x08)
        self.assertEqual(DmarParserTest.dse_t_01h.Reserved, 0)
        self.assertEqual(DmarParserTest.dse_t_01h.EnumerationID, 0x00)
        self.assertEqual(DmarParserTest.dse_t_01h.StartBusNumber, 0x00)
        self.assertEqual(DmarParserTest.dse_t_01h.Path[0], ((0x02,), (0x00,)))
        self.assertEqual(DmarParserTest.dse_t_01h.TypeString, "PCI Endpoint Device")

        try:
            dse_01_str = str(DmarParserTest.dse_t_01h)
            self.assertNotEqual(dse_01_str, None)
        except:
            self.assertFalse(False, "Failed to convert Device Scope Entry type 0 object to string")

    def test_dmar_parser_dse_03h(self):
        # IOAPIC
        dse_t_03h = bytes([0x03, 0x08, 0x00, 0x00, 0x02, 0xf0, 0x1f, 0x00])
        DmarParserTest.dse_t_03h = DMARTable.DeviceScopeStruct(dse_t_03h)
        self.assertNotEqual(DmarParserTest.dse_t_03h, None)
        self.assertEqual(DmarParserTest.dse_t_03h.Type, 0x03)
        self.assertEqual(DmarParserTest.dse_t_03h.Length, 0x08)
        self.assertEqual(DmarParserTest.dse_t_03h.Reserved, 0)
        self.assertEqual(DmarParserTest.dse_t_03h.EnumerationID, 0x02)
        self.assertEqual(DmarParserTest.dse_t_03h.StartBusNumber, 0xf0)
        self.assertEqual(DmarParserTest.dse_t_03h.Path[0], ((0x1f,), (0x00,)))
        self.assertEqual(DmarParserTest.dse_t_03h.TypeString, "IOAPIC")

        try:
            dse_03_str = str(DmarParserTest.dse_t_03h)
            self.assertNotEqual(dse_03_str, None)
        except:
            self.assertFalse(False, "Failed to convert Device Scope Entry type 3 object to string")

    def test_dmar_parser_dse_04h(self):
        # MSI_CAPABLE_HPET
        dse_t_04h = bytes([0x04, 0x08, 0x00, 0x00, 0x00, 0x00, 0x1f, 0x00])
        DmarParserTest.dse_t_04h = DMARTable.DeviceScopeStruct(dse_t_04h)
        self.assertNotEqual(DmarParserTest.dse_t_04h, None)
        self.assertEqual(DmarParserTest.dse_t_04h.Type, 0x04)
        self.assertEqual(DmarParserTest.dse_t_04h.Length, 0x08)
        self.assertEqual(DmarParserTest.dse_t_04h.Reserved, 0)
        self.assertEqual(DmarParserTest.dse_t_04h.EnumerationID, 0x00)
        self.assertEqual(DmarParserTest.dse_t_04h.StartBusNumber, 0x00)
        self.assertEqual(DmarParserTest.dse_t_04h.Path[0], ((0x1f,), (0x00,)))
        self.assertEqual(DmarParserTest.dse_t_04h.TypeString, "MSI_CAPABLE_HPET")

        try:
            dse_04_str = str(DmarParserTest.dse_t_04h)
            self.assertNotEqual(dse_04_str, None)
        except:
            self.assertFalse(False, "Failed to convert Device Scope Entry type 4 object to string")

    def test_dmar_parser_drhd(self):
        # DRHD header
        drhd_t = bytes([0x00, 0x00,         # Type
                        0x10, 0x00,         # Length
                        0x00,               # Flags
                        0x00,               # Reserved
                        0x00, 0x00,         # Segment Number
                        0x00, 0x00, 0xd9, 0xfe, 0x00, 0x00, 0x00, 0x00])    # Register Base Address

        DmarParserTest.drhd_t = DMARTable.DRHDStruct(drhd_t, len(drhd_t))
        self.assertNotEqual(DmarParserTest.drhd_t, None)

        self.assertEqual(DmarParserTest.drhd_t.Type, 0x00)
        self.assertEqual(DmarParserTest.drhd_t.Length, 0x10)
        self.assertEqual(DmarParserTest.drhd_t.Flags, 0x00)
        self.assertEqual(DmarParserTest.drhd_t.Reserved, 0x00)
        self.assertEqual(DmarParserTest.drhd_t.SegmentNumber, 0x0000)
        self.assertEqual(DmarParserTest.drhd_t.RegisterBaseAddress, 0x00000000fed90000)

        try:
            drhd_xml = DmarParserTest.drhd_t.toXml()
            self.assertNotEqual(drhd_xml, None)
        except:
            self.assertFalse(False, "Failed to convert DRHD object to xml")

        try:
            drhd_str = str(DmarParserTest.drhd_t)
            self.assertNotEqual(drhd_str, None)
        except:
            self.assertFalse(False, "Failed to convert DRHD object to string")

    def test_dmar_parser_rmrr(self):
        # RMRR header
        rmrr_t = bytes([0x01, 0x00,             # Type
                        0x18, 0x00,             # Length
                        0x00, 0x00,             # Reserved
                        0x00, 0x00,             # SegmentNumber
                        0x00, 0x00, 0x00, 0x8d, 0x00, 0x00, 0x00, 0x00,     # ReservedMemoryBaseAddress
                        0xff, 0xff, 0x7f, 0x8f, 0x00, 0x00, 0x00, 0x00])    # ReservedMemoryRegionLimitAddress

        DmarParserTest.rmrr_t = DMARTable.RMRRStruct(rmrr_t, len(rmrr_t))
        self.assertNotEqual(DmarParserTest.rmrr_t, None)

        self.assertEqual(DmarParserTest.rmrr_t.Type, 0x01)
        self.assertEqual(DmarParserTest.rmrr_t.Length, 0x18)
        self.assertEqual(DmarParserTest.rmrr_t.Reserved, 0x00)
        self.assertEqual(DmarParserTest.rmrr_t.SegmentNumber, 0x00)
        self.assertEqual(DmarParserTest.rmrr_t.ReservedMemoryBaseAddress, 0x8d000000)
        self.assertEqual(DmarParserTest.rmrr_t.ReservedMemoryRegionLimitAddress, 0x8f7fffff)

        try:
            rmrr_xml = DmarParserTest.rmrr_t.toXml()
            self.assertNotEqual(rmrr_xml, None)
        except:
            self.assertFalse(False, "Failed to convert RMRR object to xml")

        try:
            rmrr_str = str(DmarParserTest.rmrr_t)
            self.assertNotEqual(rmrr_str, None)
        except:
            self.assertFalse(False, "Failed to convert RMRR object to string")

    def test_dmar_parser_acpi_header(self):
        # Just ACPI header only
        dmar_header = bytes([0x44, 0x4d, 0x41, 0x52,    # Signature: DMAR
                            0x30, 0x00, 0x00, 0x00,     # Length
                            0x01,                       # Revision
                            0xad,                       # Checksum
                            0x4d, 0x53, 0x46, 0x54, 0x20, 0x20,     # OEM ID: 'MSFT  '
                            0x4d, 0x53, 0x46, 0x54, 0x20, 0x20, 0x20, 0x20,     # OEM Table ID: 'MSFT    '
                            0x01, 0x00, 0x00, 0x00,     # OEM Revision
                            0x49, 0x4e, 0x54, 0x4c,     # CreatorID: 'INTL'
                            0x01, 0x00, 0x00, 0x00,     # Creator revision
                            0x26,                       # HostAddressWidth
                            0x05,                       # Flags
                            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00])  # Reserved

        acpi_header = DMARTable.AcpiTableHeader(dmar_header)
        self.assertNotEqual(acpi_header, None)

        self.assertEqual(acpi_header.Signature, b'DMAR')
        self.assertEqual(acpi_header.Length, 0x30)
        self.assertEqual(acpi_header.Revision, 0x01)
        self.assertEqual(acpi_header.Checksum, 0xad)
        self.assertEqual(acpi_header.OEMID, b'MSFT  ')
        self.assertEqual(acpi_header.OEMTableID, b'MSFT    ')
        self.assertEqual(acpi_header.OEMRevision, 1)
        self.assertEqual(acpi_header.CreatorID, b'INTL')
        self.assertEqual(acpi_header.CreatorRevision, 1)
        self.assertEqual(acpi_header.HostAddressWidth, 0x26)
        self.assertEqual(acpi_header.Flags, 0x05)

        try:
            dmar_xml = acpi_header.toXml()
            self.assertNotEqual(dmar_xml, None)
        except:
            self.assertFalse(False, "Failed to convert ACPI header object to xml")

        try:
            dmar_str = str(acpi_header)
            self.assertNotEqual(dmar_str, None)
        except:
            self.assertFalse(False, "Failed to convert ACPI header object to string")

    def test_dmar_parser_ivrs_full(self):
        # finally a real deal, just to see if they are stitched properly
        full_table = bytes([0x44, 0x4d, 0x41, 0x52, 0x88, 0x00, 0x00, 0x00,
                            0x01, 0xad, 0x4d, 0x53, 0x46, 0x54, 0x20, 0x20,
                            0x4d, 0x53, 0x46, 0x54, 0x20, 0x20, 0x20, 0x20,
                            0x01, 0x00, 0x00, 0x00, 0x49, 0x4e, 0x54, 0x4c,
                            0x01, 0x00, 0x00, 0x00, 0x26, 0x05, 0x00, 0x00,
                            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                            0x00, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00,
                            0x00, 0x00, 0xd9, 0xfe, 0x00, 0x00, 0x00, 0x00,
                            0x01, 0x08, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00,
                            0x00, 0x00, 0x20, 0x00, 0x01, 0x00, 0x00, 0x00,
                            0x00, 0x10, 0xd9, 0xfe, 0x00, 0x00, 0x00, 0x00,
                            0x03, 0x08, 0x00, 0x00, 0x02, 0xf0, 0x1f, 0x00,
                            0x04, 0x08, 0x00, 0x00, 0x00, 0x00, 0x1f, 0x00,
                            0x01, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00,
                            0x00, 0x00, 0x00, 0x8d, 0x00, 0x00, 0x00, 0x00,
                            0xff, 0xff, 0x7f, 0x8f, 0x00, 0x00, 0x00, 0x00,
                            0x01, 0x08, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00])

        dmar_table = DMARTable(full_table)
        self.assertEqual(dmar_table.ANDDCount(), 0)
        self.assertTrue(dmar_table.DMARBitEnabled())
        self.assertEqual(len(dmar_table.dmar_table.RMRRlist), 1)
        self.assertFalse(dmar_table.CheckRMRRCount())
        self.assertEqual(len(dmar_table.dmar_table.SubStructs), 3)

        try:
            dmar_xml = dmar_table.toXml()
            self.assertNotEqual(dmar_xml, None)
        except:
            self.assertFalse(False, "Failed to convert DMAR object to xml")

        try:
            dmar_str = str(dmar_table)
            self.assertNotEqual(dmar_str, None)
        except:
            self.assertFalse(False, "Failed to convert DMAR object to string")


if __name__ == '__main__':
    unittest.main()
